using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace ID3TagEditLite {
    public partial class FormEditDlg : Form {
        TTagRecord record;
        public Point localLocation;
        public Point parentDesktopLocation;
        bool bMulti = false;
        public string filePath = "";
        public string lastModControlName = "";

        public FormEditDlg() {
            InitializeComponent();
            if (Program.thePref.applyFontToDialog) {
                Font = new Font(Program.thePref.fontFamily, Program.thePref.fontSize);
            }
            else {
                Font = System.Drawing.SystemFonts.MessageBoxFont;
            }
            cbxGenre.BeginUpdate();
            foreach (TGenreItem g in Program.theGenre.genres.Values) {
                cbxGenre.Items.Add(g.GetNameForMenu());
            }
            cbxGenre.EndUpdate();
        }

        public void Setup(TTagRecord record, string filePath, bool bMulti) {
            this.record = record;
            this.filePath = filePath;
            this.bMulti = bMulti;
            string lastModControlNameSaved = this.lastModControlName;

            cbxSaveEnc.Items.AddRange(TTagRecord.GetEncodingNames());
            cbxVersion.Items.AddRange(TTagRecord.GetVersionNames());

            if (this.bMulti) {
                edtFileName.Enabled = false;
                edtTitle.Enabled = false;
                edtComment.Enabled = false;
                edtTrack.Enabled = false;
            }
            else {
                edtFilePath.Text = this.filePath;
                edtFileName.Text = this.record.fileName;
                edtTitle.Text = this.record.title.str;
                edtComment.Text = this.record.comment.str;
                edtTrack.Text = this.record.GetTrackString();
            }

            edtAlbum.Text = this.record.album.str;
            edtArtist.Text = this.record.artist.str;
            edtTrackMax.Text = this.record.GetTrackMaxString();
            edtYear.Text = this.record.year;
            edtGenre.Text = this.record.GetGenreNumberString();

            TGenreItem g = Program.theGenre.GetGenreItemByGenreNumber(this.record.genre);
            if (g != null) {
                cbxGenre.SelectedIndex = Program.theGenre.genres.Values.IndexOf(g);
            }
            edtLoadEnc.Text = TTagRecord.GetEncodingString(this.record.loadEnc);
            cbxSaveEnc.SelectedIndex = (int)this.record.saveEnc;
            cbxVersion.SelectedIndex = (int)this.record.version;

            this.lastModControlName = lastModControlNameSaved;
        }

        private void FormEditDlg_Shown(object sender, EventArgs e) {
            if (lastModControlName.Length > 0) {
                Control[] c = this.Controls.Find(lastModControlName, true);
                if (c != null && c.Length > 0)
                    c[0].Focus();
            }
        }

        public TTagRecord GetTagRecord() {
            return record;
        }

        private void btnOK_Click(object sender, EventArgs e) {
            if (bMulti) {
            }
            else {
                record.SetPropertyOfFileName(edtFileName.Text);
                record.SetPropertyOfTitle(edtTitle.Text);
                record.SetPropertyOfComment(edtComment.Text);
                record.SetPropertyOfTrack(edtTrack.Text);
            }
            record.SetPropertyOfAlbum(edtAlbum.Text);
            record.SetPropertyOfArtist(edtArtist.Text);
            record.SetPropertyOfGenre(edtGenre.Text);
            record.SetPropertyOfTrackMax(edtTrackMax.Text);
            record.SetPropertyOfYear(edtYear.Text);
            record.SetPropertyOfSaveEnc((TTagEncoding)cbxSaveEnc.SelectedIndex);
            record.SetPropertyOfVersion((TTagVersion)cbxVersion.SelectedIndex);

            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        private void cbxGenre_SelectedValueChanged(object sender, EventArgs e) {
            int index = cbxGenre.SelectedIndex;
            if (index < 0)
                return;
            edtGenre.Text = Program.theGenre.genres.Values[index].number.ToString();
        }

        private void edtGenre_TextChanged(object sender, EventArgs e) {
            if (edtGenre.Text.Length > 0) {
                int n = 0;
                if (TTools.StringToInteger(ref n, edtGenre.Text) && n >= 0) {
                    TGenreItem g = Program.theGenre.GetGenreItemByGenreNumber(n);
                    if (g != null) {
                        cbxGenre.SelectedIndex = Program.theGenre.genres.Values.IndexOf(g);
                    }
                }
            }
            lastModControlName = edtGenre.Name;
        }

        private void EditDlg_Load(object sender, EventArgs e) {
            Point p = parentDesktopLocation;
            p.X += localLocation.X;
            p.Y += localLocation.Y;
            this.DesktopLocation = p;
        }

        private void EditDlg_Move(object sender, EventArgs e) {
        }

        private void EditDlg_FormClosed(object sender, FormClosedEventArgs e) {
            localLocation.X = this.DesktopLocation.X - parentDesktopLocation.X;
            localLocation.Y = this.DesktopLocation.Y - parentDesktopLocation.Y;
        }

        private void cbxSaveEnc_SelectedIndexChanged(object sender, EventArgs e) {
            if ((TTagEncoding)cbxSaveEnc.SelectedIndex == TTagEncoding.UTF16) {
                if ((TTagVersion)cbxVersion.SelectedIndex < TTagVersion.V_2_2) {
                    cbxVersion.SelectedIndex = (int)TTagVersion.V_2_2;
                }
            }
            lastModControlName = cbxSaveEnc.Name;
        }

        private void cbxVersion_SelectedIndexChanged(object sender, EventArgs e) {
            if ((TTagVersion)cbxVersion.SelectedIndex < TTagVersion.V_2_2) {
                cbxSaveEnc.SelectedIndex = (int)TTagEncoding.ANSI;
            }
            lastModControlName = cbxVersion.Name;
        }

        private void edtYear_TextChanged(object sender, EventArgs e) {
            lastModControlName = edtYear.Name;
        }

        private void edtFileName_TextChanged(object sender, EventArgs e) {
            lastModControlName = edtFileName.Name;
        }

        private void edtTitle_TextChanged(object sender, EventArgs e) {
            lastModControlName = edtTitle.Name;
        }

        private void edtAlbum_TextChanged(object sender, EventArgs e) {
            lastModControlName = edtAlbum.Name;
        }

        private void edtArtist_TextChanged(object sender, EventArgs e) {
            lastModControlName = edtArtist.Name;
        }

        private void edtComment_TextChanged(object sender, EventArgs e) {
            lastModControlName = edtComment.Name;
        }

        private void edtTrack_TextChanged(object sender, EventArgs e) {
            lastModControlName = edtTrack.Name;
        }

        private void edtTrackMax_TextChanged(object sender, EventArgs e) {
            lastModControlName = edtTrackMax.Name;
        }
    }
}