using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Globalization;

namespace ID3TagEditLite {
    public partial class FormGenreDlg : Form {
        public int number;
        public Point localLocation;
        public Point parentDesktopLocation;

        public FormGenreDlg() {
            InitializeComponent();
            if (Program.thePref.applyFontToDialog) {
                Font = new Font(Program.thePref.fontFamily, Program.thePref.fontSize);
            }
            else {
                Font = System.Drawing.SystemFonts.MessageBoxFont;
            }
            cbxGenre.BeginUpdate();
            foreach (TGenreItem g in Program.theGenre.genres.Values) {
                cbxGenre.Items.Add(g.GetNameForMenu());
            }
            cbxGenre.EndUpdate();
        }

        public void Setup(int genreNumber) {
            number = genreNumber;
            TGenreItem g = Program.theGenre.GetGenreItemByGenreNumber(genreNumber);
            if (g != null) {
                cbxGenre.SelectedIndex = Program.theGenre.genres.Values.IndexOf(g);
            }
            edtGenre.Text = genreNumber.ToString();
        }

        private void btnOK_Click(object sender, EventArgs e) {
            TTools.StringToInteger(ref number, edtGenre.Text);
            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        private void btnCancel_Click(object sender, EventArgs e) {
            this.DialogResult = DialogResult.Cancel;
            this.Close();
        }

        public int GetGenre() {
            return number;
        }

        private void GenreDlg_Load(object sender, EventArgs e) {
            Point p = parentDesktopLocation;
            p.X += localLocation.X;
            p.Y += localLocation.Y;
            this.DesktopLocation = p;
        }

        private void GenreDlg_FormClosed(object sender, FormClosedEventArgs e) {
            localLocation.X = this.DesktopLocation.X - parentDesktopLocation.X;
            localLocation.Y = this.DesktopLocation.Y - parentDesktopLocation.Y;
        }

        private void cbxGenre_SelectedValueChanged(object sender, EventArgs e) {
            int index = cbxGenre.SelectedIndex;
            if (index < 0)
                return;
            edtGenre.Text = Program.theGenre.genres.Values[index].number.ToString();
        }

        private void edtGenre_TextChanged(object sender, EventArgs e) {
            if (edtGenre.Text.Length > 0) {
                int n = 0;
                if (TTools.StringToInteger(ref n, edtGenre.Text) && n >= 0) {
                    TGenreItem g = Program.theGenre.GetGenreItemByGenreNumber(n);
                    if (g != null) {
                        cbxGenre.SelectedIndex = Program.theGenre.genres.Values.IndexOf(g);
                    }
                }
            }
        }

        private void FormGenreDlg_Shown(object sender, EventArgs e) {
            edtGenre.Focus();
        }
    }
}