using System;
using System.Collections.Generic;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Globalization;
using System.Collections.Specialized;
using System.Windows.Threading;

namespace ID3TagEditLite {
    public delegate void RedrawHandler(List<TTagItem> tags);
    public delegate List<TTagItem> GetSelectedTagItemsHandler();
    public delegate void SetUndoHandler(List<TTagItem> tags);
    public delegate void ProgressHandler(int percent);
    public delegate void SaveEndHandler();
    public delegate bool WorkingHandler();

    public enum UndoMode {
        NON = 0,
        UNDO = 1,
        REDO = 2
    }

    public partial class FormMain : Form {
        public TTagManager tagMgr = new TTagManager();
        public TPluginManager pluginMgr = new TPluginManager();
        public bool bFirstWindow = false;
        Rectangle defaultRect;
        Rectangle setupRect;
        public string originalFormName = "";
        public const string UNTITLED = "untitled";
        public string titleBaseName = UNTITLED;
        List<TTagItem> undoTargets = null;
        UndoMode undoMode = UndoMode.NON;
        public bool bCloseRequested = false;
        public Dispatcher dispather = null;

        public FormMain() {
            InitializeComponent();
            Font = System.Drawing.SystemFonts.MessageBoxFont;
            menuStripMain.Font = System.Windows.Forms.SystemInformation.MenuFont;
            fileToolStripMenuItem.Font = System.Windows.Forms.SystemInformation.MenuFont;
            editToolStripMenuItem.Font = System.Windows.Forms.SystemInformation.MenuFont;
            iD3ToolStripMenuItem.Font = System.Windows.Forms.SystemInformation.MenuFont;
            pluginToolStripMenuItem.Font = System.Windows.Forms.SystemInformation.MenuFont;
            windowToolStripMenuItem.Font = System.Windows.Forms.SystemInformation.MenuFont;
            helpToolStripMenuItem.Font = System.Windows.Forms.SystemInformation.MenuFont;
            originalFormName = this.Text;
            tagMgr.ProgressCallback = new ProgressHandler(ProgressCallback);
            tagMgr.RedrawCallback = new RedrawHandler(RedrawCallback);
            tagMgr.SaveEndCallback = new SaveEndHandler(SaveEndCallback);
            tagMgr.WorkingCallback = new WorkingHandler(WorkingCallback);
            defaultRect = this.DesktopBounds;
            if (Screen.PrimaryScreen.WorkingArea.Contains(defaultRect) == false)
                defaultRect = Screen.PrimaryScreen.WorkingArea;
            statusMessage.Text = "";
            pluginMgr.ProgressCallback = new ProgressHandler(ProgressCallback);
            pluginMgr.RedrawCallback = new RedrawHandler(RedrawCallback);
            pluginMgr.GetSelectedTagItemsCallback = new GetSelectedTagItemsHandler(GetSelectedTagItemsCallback);
            pluginMgr.SetUndoCallback = new SetUndoHandler(SetUndoCallback);
            pluginMgr.WorkingCallback = new WorkingHandler(WorkingCallback);
            dispather = Dispatcher.CurrentDispatcher;

            listViewMain.Columns[modify.Index].Name = "modify";
            listViewMain.Columns[track.Index].Name = "track";
            listViewMain.Columns[fileName.Index].Name = "file-name";
            listViewMain.Columns[title.Index].Name = "title";
            listViewMain.Columns[artist.Index].Name = "artist";
            listViewMain.Columns[album.Index].Name = "album";
            listViewMain.Columns[year.Index].Name = "year";
            listViewMain.Columns[comment.Index].Name = "comment";
            listViewMain.Columns[genre.Index].Name = "genre";
            listViewMain.Columns[version.Index].Name = "version";
            listViewMain.Columns[loadEncode.Index].Name = "load-encode";
            listViewMain.Columns[saveEncode.Index].Name = "save-encode";

            listViewMain.Columns[modify.Index].Width = Program.thePref.columnWidthModify;
            listViewMain.Columns[track.Index].Width = Program.thePref.columnWidthTrack;
            listViewMain.Columns[fileName.Index].Width = Program.thePref.columnWidthFileName;
            listViewMain.Columns[title.Index].Width = Program.thePref.columnWidthTitle;
            listViewMain.Columns[artist.Index].Width = Program.thePref.columnWidthArtist;
            listViewMain.Columns[album.Index].Width = Program.thePref.columnWidthAlbum;
            listViewMain.Columns[year.Index].Width = Program.thePref.columnWidthYear;
            listViewMain.Columns[comment.Index].Width = Program.thePref.columnWidthComment;
            listViewMain.Columns[genre.Index].Width = Program.thePref.columnWidthGenre;
            listViewMain.Columns[version.Index].Width = Program.thePref.columnWidthVersion;
            listViewMain.Columns[loadEncode.Index].Width = Program.thePref.columnWidthSourceEncode;
            listViewMain.Columns[saveEncode.Index].Width = Program.thePref.columnWidthTargetEncode;
            listViewMain.Font = new Font(Program.thePref.fontFamily, Program.thePref.fontSize);
        }

        private void MainForm_Load(object sender, EventArgs e) {
            if (setupRect.IsEmpty == false) {
                DesktopBounds = setupRect;
            }
            // vOC
            pluginMgr.Setup();
            foreach (TPluginItem p in pluginMgr.plugins.Values) {
                ToolStripMenuItem m = new ToolStripMenuItem(p.GetNameForMenu(), null, new EventHandler(p.OnClick));
                this.pluginToolStripMenuItem.DropDownItems.Add(m);
            }
            // GR[fBO
            saveEncAnsiToolStripMenuItem.Text = "&" + TTagRecord.GetEncodingString(TTagEncoding.ANSI);
            saveEncUnicodeToolStripMenuItem.Text = "&" + TTagRecord.GetEncodingString(TTagEncoding.UTF16);
            saveEncAnsiToolStripMenuItem.Tag = TTagEncoding.ANSI;
            saveEncUnicodeToolStripMenuItem.Tag = TTagEncoding.UTF16;
            // ^Oo[W
            v10ToolStripMenuItem.Tag = TTagVersion.V_1_0;
            v11ToolStripMenuItem.Tag = TTagVersion.V_1_1;
            v22ToolStripMenuItem.Tag = TTagVersion.V_2_2;
            v23ToolStripMenuItem.Tag = TTagVersion.V_2_3;
            v24ToolStripMenuItem.Tag = TTagVersion.V_2_4;

            ChangeTitle(titleBaseName);
            FixMenu();
            FixWindowMenu();
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs e) {
            if (bFirstWindow) {
                if (WindowState == FormWindowState.Normal)
                    Program.thePref.userBounds = this.DesktopBounds;
                Program.thePref.columnWidthModify = listViewMain.Columns[modify.Index].Width;
                Program.thePref.columnWidthTrack = listViewMain.Columns[track.Index].Width;
                Program.thePref.columnWidthFileName = listViewMain.Columns[fileName.Index].Width;
                Program.thePref.columnWidthTitle = listViewMain.Columns[title.Index].Width;
                Program.thePref.columnWidthArtist = listViewMain.Columns[artist.Index].Width;
                Program.thePref.columnWidthAlbum = listViewMain.Columns[album.Index].Width;
                Program.thePref.columnWidthYear = listViewMain.Columns[year.Index].Width;
                Program.thePref.columnWidthComment = listViewMain.Columns[comment.Index].Width;
                Program.thePref.columnWidthGenre = listViewMain.Columns[genre.Index].Width;
                Program.thePref.columnWidthVersion = listViewMain.Columns[version.Index].Width;
                Program.thePref.columnWidthSourceEncode = listViewMain.Columns[loadEncode.Index].Width;
                Program.thePref.columnWidthTargetEncode = listViewMain.Columns[saveEncode.Index].Width;
            }
            Program.DisposeFormMain(this);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e) {
            List<FormMain> list = new List<FormMain>(Program.mainFormList);
            foreach (FormMain f in list) {
                f.Close();
            }
        }

        void ChangeTitle(string title) {
            titleBaseName = title;
            this.Text = originalFormName + " : " + titleBaseName;
        }
        
        private void MainForm_DragDrop(object sender, DragEventArgs e) {
            string[] ss = (string[])e.Data.GetData(DataFormats.FileDrop, false);
            OnLoadAnyFiles(ss);
        }

        void OnLoadAnyFiles(string[] files) {
            List<string> mp3s = new List<string>();
            foreach (string path in files) {
                if (File.Exists(path) == false) {
                    if (Directory.Exists(path))
                        mp3s.Add(path);
                }
                else {
                    string ext = Path.GetExtension(path).ToLower();
                    if (ext.Equals(".mp3"))
                        mp3s.Add(path);
                    else if (ext.Equals(".tsv") || ext.Equals(".csv")) {
                        Import(path);
                        return;
                    }
                }
            }
            if (mp3s.Count > 0)
                AddMP3Files(mp3s.ToArray());
            FixMenu();
        }

        private void MainForm_DragEnter(object sender, DragEventArgs e) {
            e.Effect = e.Data.GetDataPresent(DataFormats.FileDrop) ? DragDropEffects.All : DragDropEffects.None;
        }

        string[] SelectAudioFiles() {
            openFileDialog.InitialDirectory = Program.thePref.loadDir;
            openFileDialog.Filter = "MP3 (*.mp3)|*.mp3|All files (*.*)|*.*";
            openFileDialog.DefaultExt = "mp3";
            DialogResult result = openFileDialog.ShowDialog();
            if (result == DialogResult.OK) {
                Program.thePref.loadDir = openFileDialog.InitialDirectory;
                return openFileDialog.FileNames;
            }
            return null;
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e) {
            string[] files = SelectAudioFiles();
            if (files != null && files.Length > 0) {
                FormMain f = Program.NewFormMain(1, DesktopBounds);
                f.AddMP3Files(files);
                FixWindowMenuAll();
            }
            FixMenu();
        }
        
        private void addToolStripMenuItem_Click(object sender, EventArgs e) {
            string[] files = SelectAudioFiles();
            if (files != null && files.Length > 0)
                AddMP3Files(files);
            FixMenu();
        }

        public void AddMP3Files(string[] s) {
            if (s.Length <= 0)
                return;
            bool bSort = tagMgr.tags.Count == 0;
            List<string> list = new List<string>(s);
            list.Sort();
            s = list.ToArray();
            if (tagMgr.tags.Count == 0) {
                string path = s[0];
                string name = "";
                if (File.Exists(path) == false && Directory.Exists(path))
                    name = Path.GetFileName(path);
                else
                    name = Path.GetFileName(Path.GetDirectoryName(path));
                ChangeTitle(name);
                FixWindowMenuAll();
            }
            int n = tagMgr.AppendFiles(s);
            if (bSort)
                tagMgr.tags.Sort(new TTagComparer());
            IncreaseListViewMain(n);
            RedrawListViewMain();
        }

        public void RedrawListViewMainSub(TTagItem t) {
            int n = tagMgr.tags.IndexOf(t);
            if (n < 0)
                return;
            ListViewItem item = listViewMain.Items[n];
            item.SubItems[modify.Index].Text = t.record.GetModifyString();
            item.SubItems[track.Index].Text = t.record.GetTrackString();
            item.SubItems[fileName.Index].Text = t.record.fileName;
            item.SubItems[title.Index].Text = t.record.title.str;
            item.SubItems[artist.Index].Text = t.record.artist.str;
            item.SubItems[album.Index].Text = t.record.album.str;
            item.SubItems[year.Index].Text = t.record.year;
            item.SubItems[comment.Index].Text = t.record.comment.str;
            item.SubItems[genre.Index].Text = t.record.GetGenreString();
            item.SubItems[version.Index].Text = t.record.GetVersionString();
            item.SubItems[loadEncode.Index].Text = TTagRecord.GetEncodingString(t.record.loadEnc);
            item.SubItems[saveEncode.Index].Text = TTagRecord.GetEncodingString(t.record.saveEnc);
        }

        public void ClearListViewMainSub(int n) {
            ListViewItem item = listViewMain.Items[n];
            item.SubItems[modify.Index].Text = "";
            item.SubItems[track.Index].Text = "";
            item.SubItems[fileName.Index].Text = "";
            item.SubItems[title.Index].Text = "";
            item.SubItems[artist.Index].Text = "";
            item.SubItems[album.Index].Text = "";
            item.SubItems[year.Index].Text = "";
            item.SubItems[comment.Index].Text = "";
            item.SubItems[genre.Index].Text = "";
            item.SubItems[version.Index].Text = "";
            item.SubItems[loadEncode.Index].Text = "";
            item.SubItems[saveEncode.Index].Text = "";
        }

        public void RedrawListViewMain(TTagItem t) {
            listViewMain.BeginUpdate();
            RedrawListViewMainSub(t);
            listViewMain.EndUpdate();
            FixMenu();
        }

        public void RedrawListViewMain(List<TTagItem> tags) {
            listViewMain.BeginUpdate();
            foreach (TTagItem t in tags) {
                RedrawListViewMainSub(t);
            }
            listViewMain.EndUpdate();
            FixMenu();
        }

        public void RedrawListViewMain() {
            RedrawListViewMain(tagMgr.tags);
        }

        void IncreaseListViewMain(int n) {
            for (int i = 0; i < n; i++) {
                ListViewItem item = new ListViewItem("");
                for (int j = 1; j < listViewMain.Columns.Count; j++) {
                    item.SubItems.Add("");
                }
                listViewMain.Items.Add(item);
            }
        }

        private void listViewMain_DoubleClick(object sender, EventArgs e) {
            editTagToolStripMenuItem_Click(sender, e);
        }

        string lastModControlName = "";

        void EditWithDlg(List<TTagItem> tags) {
            if (tags == null || tags.Count <= 0)
                return;
            FormEditDlg dlg = new FormEditDlg();
            dlg.lastModControlName = lastModControlName;
            dlg.localLocation = Program.thePref.editDlgLocation;
            dlg.parentDesktopLocation = this.DesktopLocation;
            dlg.Setup(tags[0].record.Clone(), tags.Count == 1 ? tags[0].filePath : "", tags.Count > 1);
            if (dlg.ShowDialog() == DialogResult.OK) {
                TTagRecord r = dlg.GetTagRecord();
                SetUndo(tags);
                if (tags.Count > 1) {
                    foreach (TTagItem t in tags) {
                        t.record.SetPropertyOfAlbum(r.GetAlbum());
                        t.record.SetPropertyOfGenre(r.GetGenreNumberString());
                        t.record.SetPropertyOfArtist(r.GetArtist());
                        t.record.SetPropertyOfTrackMax(r.GetTrackMaxString());
                        t.record.SetPropertyOfYear(r.GetYear());
                        t.record.SetPropertyOfLoadEnc(r.loadEnc);
                        t.record.SetPropertyOfSaveEnc(r.saveEnc);
                        t.record.SetPropertyOfVersion(r.version);
                    }
                }
                else {
                    tags[0].record = r;
                }
                RedrawListViewMain(tags);
            }
            lastModControlName = dlg.lastModControlName;
            Program.thePref.editDlgLocation = dlg.localLocation;
        }
        
        private void editTagToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags == null || tags.Count <= 0)
                return;
            EditWithDlg(tags);
            FixMenu();
        }

        private void editAllToolStripMenuItem_Click(object sender, EventArgs e) {
            if (tagMgr.tags.Count <= 0)
                return;
            EditWithDlg(tagMgr.tags);
            FixMenu();
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e) {
            tagMgr.SaveFiles();
            FixMenu();
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e) {
            FormMain f = Program.NewFormMain(1, DesktopBounds);
            FixWindowMenuAll();
            FixMenu();
        }

        private void closeToolStripMenuItem_Click(object sender, EventArgs e) {
            this.Close();
        }

        public void SetupNewWindowRect(int n, Rectangle userRect) {
            this.setupRect = CalcNewWindowRect(n, userRect);
            this.Visible = true;
        }

        public Rectangle CalcNewWindowRect(int gap, Rectangle userRect) {
            int gapx = 20;
            int gapy = 20;
            Rectangle desktop = Screen.PrimaryScreen.WorkingArea;
            /// ʂOĂ
            if (desktop.Contains(userRect) == false) {
                return defaultRect;
            }
            Rectangle r = userRect;
            r.Offset(gapx * gap, gapy * gap);
            // ʂɓĂ
            if (desktop.Contains(r)) {
                return r;
            }
            // ړʉʂnYꍇ
            // ړ\
            int m;
            r = userRect;
            for (m = 0;; m++) {
                r.Offset(gapx, gapy);
                if (desktop.Contains(r) == false)
                    break;
            }
            gap = gap % (m + 1);
            r = userRect;
            r.Offset(gapx * gap, gapy * gap);
            return r;
        }

        private void MainForm_Move(object sender, EventArgs e) {
        }

        private void MainForm_SizeChanged(object sender, EventArgs e) {
            if (bFirstWindow)
                if (WindowState == FormWindowState.Normal)
                    Program.thePref.userBounds = this.DesktopBounds;
        }

        string GetExportExtByIndex(int n) {
            if (n == 2)
                return "csv";
            return "tsv";
        }

        private DialogResult StopSaveConfirm() {
            return MessageBox.Show(this, "Do You Stop Save? \"" + titleBaseName + "\"", "Confirm", MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);
        }

        private DialogResult SaveConfirm() {
            return MessageBox.Show(this, "Do You Save? \"" + titleBaseName + "\"", "Confirm", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e) {
            if (pluginMgr.IsWorking()) {
                pluginMgr.Stop();
            }
            if (tagMgr.IsWorking()) {
                DialogResult result = StopSaveConfirm();
                if (result == DialogResult.No) {
                    e.Cancel = true;
                    FixMenu();
                }
                else if (result == DialogResult.Yes) {
                    tagMgr.Stop();
                }
            }
            else if (tagMgr.IsModified()) {
                DialogResult result = SaveConfirm();
                if (result == DialogResult.Cancel) {
                    e.Cancel = true;
                    FixMenu();
                }
                else if (result == DialogResult.Yes) {
                    tagMgr.SaveFiles();
                    e.Cancel = true;
                    bCloseRequested = true;
                    FixMenu();
                }
            }
        }

        public List<TTagItem> GetSelectedTagItems() {
            IEnumerator en = listViewMain.SelectedItems.GetEnumerator();
            List<TTagItem> list = new List<TTagItem>();
            while (en.MoveNext()) {
                ListViewItem item = (ListViewItem)en.Current;
                int n = listViewMain.Items.IndexOf(item);
                list.Add(tagMgr.tags[n]);
            }
            if (list.Count == 0 && Program.thePref.nothingIsAll) {
                return tagMgr.tags;
            }
            return list;
        }

        public void SetUndo(List<TTagItem> tags) {
            undoTargets = tags;
            foreach (TTagItem t in undoTargets) {
                t.CopyToUndoRecord();
            }
            undoMode = UndoMode.UNDO;
        }

        private void undoToolStripMenuItem_Click(object sender, EventArgs e) {
            if (undoMode == UndoMode.UNDO) {
                foreach (TTagItem t in undoTargets) {
                    t.CopyToRedoRecord();
                    t.CopyFromUndoRecord();
                }
                undoMode = UndoMode.REDO;
                undoToolStripMenuItem.Text = "Redo";
            }
            else if (undoMode == UndoMode.REDO) {
                foreach (TTagItem t in undoTargets) {
                    t.CopyToUndoRecord();
                    t.CopyFromRedoRecord();
                }
                undoMode = UndoMode.UNDO;
                undoToolStripMenuItem.Text = "Undo";
            }
            RedrawListViewMain(undoTargets);
        }

        private void listViewMain_ColumnClick(object sender, ColumnClickEventArgs e) {
            tagMgr.sortMethod = listViewMain.Columns[e.Column].Name;
            tagMgr.tags.Sort(new TTagComparer());
            RedrawListViewMain();
        }

        private void autoTrackNumberToolStripMenuItem_Click(object sender, EventArgs e) {
            int i = 1;
            foreach (TTagItem t in tagMgr.tags) {
                t.record.SetPropertyOfTrack(i.ToString());
                i++;
            }
            RedrawListViewMain();
        }

        private void copyToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags == null || tags.Count == 0)
                return;
            string s = "";
            if (Program.thePref.copyOnlyTitleAndArtist) {
                foreach (TTagItem t in tags) {
                    s += t.record.GetTitle() + " " + t.record.GetArtist();
                    s += "\n";
                }
            }
            else {
                foreach (TTagItem t in tags) {
                    s += t.record.GetCSVLine();
                    s += "\n";
                }
            }
            try {
                Clipboard.Clear();
                //Clipboard.SetText(s, TextDataFormat.Text);
                //Clipboard.SetText(s, TextDataFormat.CommaSeparatedValue);
                Clipboard.SetText(s, TextDataFormat.UnicodeText);
            }
            catch (Exception ex) {
                ///////////////////////////////////////
                // 
                statusMessage.Text = ex.Message;
            }
            FixMenu();
        }

        private void pasteToolStripMenuItem_Click(object sender, EventArgs e) {
            if (Clipboard.ContainsText()) {
                OnPasteText();
            }
            else if (Clipboard.ContainsFileDropList()) {
                StringCollection sc = Clipboard.GetFileDropList();
                string[] ss = new string[sc.Count];
                for (int i = 0; i < sc.Count; i++) {
                    ss[i] = sc[i];
                }
                OnLoadAnyFiles(ss);
            }
        }

        void OnPasteText() {
            string clip = Clipboard.GetText();
            char[] sep = new char[] { };
            string[] lines = clip.Split(new char[] { '\n', '\r' }, StringSplitOptions.RemoveEmptyEntries);
            if (lines == null || lines.Length == 0)
                return;

            int n = 0;
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags.Count > 0) {
                n = tagMgr.tags.IndexOf(tags[0]);
            }
            List<TTagItem> target = new List<TTagItem>();
            for (int i = 0; i < lines.Length && n < tagMgr.tags.Count; i++, n++) {
                target.Add(tagMgr.tags[n]);
            }
            if (target.Count == 0)
                return;
            SetUndo(target);
            for (int i = 0; i < target.Count; i++) {
                target[i].record.SetCSVLine(lines[i]);
            }
            RedrawListViewMain(target);
        }

        private void selectAllToolStripMenuItem_Click(object sender, EventArgs e) {
            for (int i = 0; i < listViewMain.Items.Count; i++) {
                listViewMain.Items[i].Selected = true;
            }
            FixMenu();
        }

        private void statusStripMain_ItemClicked(object sender, ToolStripItemClickedEventArgs e) {
            statusMessage.Text = "";
            FixMenu();
        }

        private void exportToolStripMenuItem_Click(object sender, EventArgs e) {
            if (tagMgr.tags.Count == 0)
                return;
            exportFileDialog.InitialDirectory = Program.thePref.exportDir;
            exportFileDialog.Filter = "TSV (*.tsv)|*.tsv|CSV (*.csv)|*.csv|All files (*.*)|*.*";
            exportFileDialog.FilterIndex = Program.thePref.exportIndex;
            exportFileDialog.DefaultExt = GetExportExtByIndex(exportFileDialog.FilterIndex);
            exportFileDialog.FileName = tagMgr.tags[0].record.GetAlbum();
            if (exportFileDialog.FileName.Length == 0)
                exportFileDialog.FileName = titleBaseName;
            //exportFileDialog.FileName += ("." + exportFileDialog.DefaultExt);
            DialogResult result;
            try {
                result = exportFileDialog.ShowDialog();
            }
            catch {
                exportFileDialog.FileName = titleBaseName;
                result = exportFileDialog.ShowDialog();
            }
            if (result == DialogResult.OK) {
                Program.thePref.exportIndex = exportFileDialog.FilterIndex;
                Program.thePref.exportDir = exportFileDialog.InitialDirectory;
                string ext = Path.GetExtension(exportFileDialog.FileName).ToLower();
                tagMgr.Export(exportFileDialog.FileName, ext);
            }
            FixMenu();
        }

        private void impotToolStripMenuItem_Click(object sender, EventArgs e) {
            if (tagMgr.tags.Count == 0)
                return;
            importFileDialog.InitialDirectory = Program.thePref.importDir;
            importFileDialog.Filter = "TSV (*.tsv)|*.tsv|CSV (*.csv)|*.csv|All files (*.*)|*.*";
            importFileDialog.FilterIndex = Program.thePref.importIndex;
            importFileDialog.DefaultExt = GetExportExtByIndex(importFileDialog.FilterIndex);
            //importFileDialog.FileName = m_titleBaseName /*+ "." + saveFileDialog.DefaultExt*/;
            DialogResult result = importFileDialog.ShowDialog();
            if (result == DialogResult.OK) {
                Program.thePref.importIndex = importFileDialog.FilterIndex;
                Program.thePref.importDir = importFileDialog.InitialDirectory;
                Import(importFileDialog.FileName);
            }
            FixMenu();
        }

        void Import(string filePath) {
            string ext = Path.GetExtension(filePath).ToLower();
            SetUndo(tagMgr.tags);
            tagMgr.Import(filePath, ext);
            RedrawListViewMain();
        }

        private void saveEncAnsiToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.SetPropertyOfSaveEnc(TTagEncoding.ANSI);
            }
            RedrawListViewMain(tags);
        }

        private void saveEncUnicodeToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.SetPropertyOfSaveEnc(TTagEncoding.UTF16);
                if (t.record.version < TTagVersion.V_2_2)
                    t.record.SetPropertyOfVersion(TTagVersion.V_2_2);
            }
            RedrawListViewMain(tags);
        }

        private void v10ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.SetPropertyOfVersion(TTagVersion.V_1_0);
                t.record.SetPropertyOfSaveEnc(TTagEncoding.ANSI);
            }
            RedrawListViewMain(tags);
        }

        private void v11ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.SetPropertyOfVersion(TTagVersion.V_1_1);
                t.record.SetPropertyOfSaveEnc(TTagEncoding.ANSI);
            }
            RedrawListViewMain(tags);
        }

        private void v22ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.SetPropertyOfVersion(TTagVersion.V_2_2);
            }
            RedrawListViewMain(tags);
        }

        private void v23ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.SetPropertyOfVersion(TTagVersion.V_2_3);
            }
            RedrawListViewMain(tags);
        }

        private void v24ToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.SetPropertyOfVersion(TTagVersion.V_2_4);
            }
            RedrawListViewMain(tags);
        }

        private void OnSelectGenre(object sender, EventArgs e) {
            ToolStripMenuItem m = (ToolStripMenuItem)sender;
            TGenreItem g = (TGenreItem)m.Tag;
            List<TTagItem> tags = GetSelectedTagItems();
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.SetPropertyOfGenre(g.number.ToString());
            }
            RedrawListViewMain(tags);
        }

        private void OnSelectWindow(object sender, EventArgs e) {
            FormMain w = (FormMain)((ToolStripMenuItem)sender).Tag;
            w.BringToFront();
            FixMenu();
        }

        private void genreToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags.Count == 0)
                return;
            FormGenreDlg dlg = new FormGenreDlg();
            dlg.parentDesktopLocation = this.DesktopLocation;
            dlg.localLocation = Program.thePref.genreDlgLocation;
            dlg.Setup(tags[0].record.genre);
            if (dlg.ShowDialog() == DialogResult.OK) {
                SetUndo(tags);
                foreach (TTagItem t in tags) {
                    t.record.SetPropertyOfGenre(dlg.GetGenre().ToString());
                }
                RedrawListViewMain(tags);
            }
            Program.thePref.genreDlgLocation = dlg.localLocation;
            FixMenu();
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e) {
            FormAboutDlg dlg = new FormAboutDlg();
            dlg.ShowDialog();
            FixMenu();
        }

        public void FixWindowMenu() {
            windowToolStripMenuItem.DropDownItems.Clear();
            foreach (FormMain w in Program.mainFormList) {
                ToolStripMenuItem m = new ToolStripMenuItem(w.titleBaseName, null, new EventHandler(OnSelectWindow));
                m.Tag = w;
                windowToolStripMenuItem.DropDownItems.Add(m);
            }
        }

        public void FixWindowMenuAll() {
            foreach (FormMain w in Program.mainFormList) {
                w.FixWindowMenu();
            }
        }

        private void preferencesToolStripMenuItem_Click(object sender, EventArgs e) {
            FormPreferencesDlg dlg = new FormPreferencesDlg();
            if (dlg.ShowDialog() == DialogResult.OK) {
                foreach (FormMain f in Program.mainFormList) {
                    foreach (TTagItem t in f.tagMgr.tags) {
                        t.record.DecodeStrings();
                    }
                    f.RedrawListViewMain();
                }
            }
        }

        public void ProgressCallback(int percent) {
            dispather.Invoke(new Action(() => {
                toolStripProgressBar.Value = percent;
            }));
        }

        public void SaveEndCallback() {
            dispather.Invoke(new Action(() => {
                toolStripProgressBar.Value = 0;
                if (bCloseRequested) {
                    bCloseRequested = false;
                    Close();
                }
            }));
        }

        public void RedrawCallback(List<TTagItem> tags) {
            dispather.Invoke(new Action(() => {
                RedrawListViewMain(tags);
            }));
        }

        public List<TTagItem> GetSelectedTagItemsCallback() {
            return GetSelectedTagItems();
        }

        public void SetUndoCallback(List<TTagItem> tags) {
            SetUndo(tags);
        }

        public bool WorkingCallback() {
            return tagMgr.IsWorking() || pluginMgr.IsWorking();
        }

        private void listViewMain_SelectedIndexChanged(object sender, EventArgs e) {
            FixMenu();
        }

        private void menuStripMain_ItemClicked(object sender, ToolStripItemClickedEventArgs e) {
            FixMenu();
        }

        void FixMenu() {
            if (undoTargets == null || undoTargets.Count == 0)
                undoToolStripMenuItem.Enabled = false;
            else
                undoToolStripMenuItem.Enabled = true;
            if (tagMgr.tags.Count == 0) {
                copyToolStripMenuItem.Enabled = false;
                iD3ToolStripMenuItem.Enabled = false;
                pluginToolStripMenuItem.Enabled = false;
                saveToolStripMenuItem.Enabled = false;
                exportToolStripMenuItem.Enabled = false;
                impotToolStripMenuItem.Enabled = false;
            }
            else {
                iD3ToolStripMenuItem.Enabled = true;
                editAllToolStripMenuItem.Enabled = true;
                saveToolStripMenuItem.Enabled = true;
                exportToolStripMenuItem.Enabled = true;
                impotToolStripMenuItem.Enabled = true;
                List<TTagItem> tags = GetSelectedTagItems();
                if (tags.Count == 0) {
                    copyToolStripMenuItem.Enabled = false;
                    editTagToolStripMenuItem.Enabled = false;
                    pluginToolStripMenuItem.Enabled = false;
                    saveEncToolStripMenuItem.Enabled = false;
                    versionToolStripMenuItem.Enabled = false;
                    genreToolStripMenuItem.Enabled = false;
                    touchToolStripMenuItem.Enabled = false;
                    appendCoverImageToolStripMenuItem.Enabled = false;
                    removeCoverImageToolStripMenuItem.Enabled = false;
                }
                else {
                    copyToolStripMenuItem.Enabled = true;
                    editTagToolStripMenuItem.Enabled = true;
                    pluginToolStripMenuItem.Enabled = true;
                    saveEncToolStripMenuItem.Enabled = true;
                    versionToolStripMenuItem.Enabled = true;
                    genreToolStripMenuItem.Enabled = true;
                    touchToolStripMenuItem.Enabled = true;
                    appendCoverImageToolStripMenuItem.Enabled = true;
                    foreach (TTagItem t in tags) {
                        if (!t.record.picture.IsNull()) {
                            removeCoverImageToolStripMenuItem.Enabled = true;
                            break;
                        }
                    }
                    saveEncAnsiToolStripMenuItem.Text = "&" + TTagRecord.GetEncodingString(TTagEncoding.ANSI);
                    saveEncUnicodeToolStripMenuItem.Text = "&" + TTagRecord.GetEncodingString(TTagEncoding.UTF16);
                    if (tags.Count == 1) {
                        TTagItem t = tags[0];
                        foreach (ToolStripMenuItem m in saveEncToolStripMenuItem.DropDownItems) {
                            m.CheckState = t.record.saveEnc == (TTagEncoding)m.Tag ? CheckState.Checked : CheckState.Unchecked;
                        }
                        foreach (ToolStripMenuItem m in versionToolStripMenuItem.DropDownItems) {
                            m.CheckState = t.record.version == (TTagVersion)m.Tag ? CheckState.Checked : CheckState.Unchecked;
                        }
                        foreach (ToolStripMenuItem m in genreToolStripMenuItem.DropDownItems) {
                            m.CheckState = t.record.genre == ((TGenreItem)m.Tag).number ? CheckState.Checked : CheckState.Unchecked;
                        }
                    }
                    else {
                        foreach (ToolStripMenuItem m in saveEncToolStripMenuItem.DropDownItems) {
                            m.CheckState = CheckState.Unchecked;
                        }
                        foreach (ToolStripMenuItem m in versionToolStripMenuItem.DropDownItems) {
                            m.CheckState = CheckState.Unchecked;
                        }
                        foreach (ToolStripMenuItem m in genreToolStripMenuItem.DropDownItems) {
                            m.CheckState = CheckState.Unchecked;
                        }
                    }
                }
            }
            if (Clipboard.ContainsText() || Clipboard.ContainsFileDropList()) {
                pasteToolStripMenuItem.Enabled = true;
            }
            else {
                pasteToolStripMenuItem.Enabled = false;
            }
            redoPluginToolStripMenuItem.Enabled = pluginMgr.redoName.Length > 0;
        }

        private void listViewMain_KeyUp(object sender, KeyEventArgs e) {
        }

        private void listViewMain_KeyPress(object sender, KeyPressEventArgs e) {
        }

        private void listViewMain_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode == Keys.Enter) {
                if (e.Shift) {
                    editAllToolStripMenuItem_Click(sender, e);
                }
                else {
                    editTagToolStripMenuItem_Click(sender, e);
                }
            }
            else if (e.KeyCode == Keys.Escape) {
                if (pluginMgr.IsWorking()) {
                    pluginMgr.Stop();
                }
                else if (tagMgr.IsWorking()) {
                    tagMgr.Stop();
                }
                else { 
                    foreach (ListViewItem item in listViewMain.Items) {
                        item.Selected = false;
                    }
                }
                FixMenu();
            }
        }

        private void touchToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags.Count == 0) {
                return;
            }
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                t.record.bModify = true;
            }
            RedrawListViewMain(tags);
            FixMenu();
        }

        private void pluginToolStripMenuItem_Click(object sender, EventArgs e) {
            FixMenu();
        }

        private void appendCoverImageToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags.Count == 0) {
                return;
            }
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                if (t.record.picture.LoadCoverFile(t.filePath)) {
                    t.record.bModify = true;
                }
            }
            RedrawListViewMain(tags);
            FixMenu();
        }

        private void removeCoverImageToolStripMenuItem_Click(object sender, EventArgs e) {
            List<TTagItem> tags = GetSelectedTagItems();
            if (tags.Count == 0) {
                return;
            }
            SetUndo(tags);
            foreach (TTagItem t in tags) {
                if (!t.record.picture.IsNull()) {
                    t.record.picture.Init();
                    while (t.record.otherPicture.Count > 0)
                        t.record.otherPicture.RemoveAt(0);
                    t.record.bModify = true;
                }
            }
            RedrawListViewMain(tags);
            FixMenu();
        }

        private void clearFileListToolStripMenuItem_Click(object sender, EventArgs e) {
            if (pluginMgr.IsWorking()) {
                pluginMgr.Stop();
            }
            if (tagMgr.IsWorking()) {
                DialogResult result = StopSaveConfirm();
                if (result == DialogResult.No) {
                    return;
                }
                else if (result == DialogResult.Yes) {
                    tagMgr.Stop();
                }
            }
            else if (tagMgr.IsModified()) {
                DialogResult result = SaveConfirm();
                if (result == DialogResult.Cancel) {
                    return;
                }
                else if (result == DialogResult.Yes) {
                    tagMgr.SaveFiles();
                    return;
                }
            }
            while (tagMgr.tags.Count > 0) {
                tagMgr.tags.RemoveAt(0);
            }
            listViewMain.Items.Clear();
            ChangeTitle(UNTITLED);
            FixMenu();
        }

        private void redoPluginToolStripMenuItem_Click(object sender, EventArgs e) {
            pluginMgr.Redo();
            FixMenu();
        }

        private void fontToolStripMenuItem_Click(object sender, EventArgs e) {
            fontDialog.Font = listViewMain.Font;
            fontDialog.ShowEffects = false;
            fontDialog.ShowColor = false;
            if (fontDialog.ShowDialog() != DialogResult.OK) {
                return;
            }
            Program.thePref.fontFamily = fontDialog.Font.FontFamily.Name;
            Program.thePref.fontSize = fontDialog.Font.Size;
            Font font = new Font(Program.thePref.fontFamily, Program.thePref.fontSize);
            foreach (FormMain mf in Program.mainFormList) {
                mf.listViewMain.Font = font;
            }
        }
    }
}
