using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace ID3TagEditLite {

    public class TID3V2Comment {
        public string language = "eng";
        public TTagString description = new TTagString();
        public TTagString text = new TTagString();

        public bool DecodeString(TTagEncoding code) {
            description.DecodeString(/*code*/);
            return text.DecodeString(/*code*/);
        }

        public void EncodeString(TTagEncoding code, int limit) {
            description.EncodeString(code, limit);
            text.EncodeString(code, limit);
        }

        public TID3V2Comment Clone() {
            TID3V2Comment c = new TID3V2Comment();
            c.language = (string)language.Clone();
            c.description = description.Clone();
            c.text = text.Clone();
            return c;
        }
    }

    public class TID3V2Head {
        /*
        struct ID3_V2_HEAD {
            char m_tag[3];
            char m_major;
            char m_minor;
            char m_flags;
            char m_size[4];
        };
        */
        public const int HEAD_SIZE = 10;
        public const int FOOT_SIZE = HEAD_SIZE;
        public const int TAG_V2_HEAD_FLAG_FOOTER = 0x10;
        public const int TAG_V2_HEAD_FLAG_UNSYNC = 0x80;
        public const int TAG_V2_HEAD_FLAG_EXTENDED = 0x40;
        public const int TAG_V2_HEAD_FLAG_EXPERIMENTAL = 0x20;
        //
        public byte[] magic;
        public byte major = (byte)TTagVersion.V_2_2;
        public byte minor = 0;
        public byte flags = 0;
        public int size = 0;

        public bool Read(BinaryReader br) {
            magic = br.ReadBytes(3);
            ASCIIEncoding ae = new ASCIIEncoding();
            if (ae.GetString(magic).Equals("ID3") == false)
                return false;
            major = br.ReadByte();
            minor = br.ReadByte();
            flags = br.ReadByte();
            byte[] data = br.ReadBytes(4);
            size = TagSizeToInteger(data);
            if (major == (byte)TTagVersion.V_2_4 && (flags & TAG_V2_HEAD_FLAG_FOOTER) != 0)
                size += FOOT_SIZE;
            return true;
        }

        public bool Write(BinaryWriter bw) {
            ASCIIEncoding ae = new ASCIIEncoding();
            magic = ae.GetBytes("ID3");
            bw.Write(magic);
            bw.Write(major);
            bw.Write(minor);
            bw.Write(flags);
            byte[] sz = TagSizeFromInteger(size);
            bw.Write(sz);
            return true;
        }

        int TagSizeToInteger(byte[] data) {
            int n = (data[0] << (7 * 3)) + (data[1] << (7 * 2)) + (data[2] << (7 * 1)) + (data[3]);
            return n;
        }

        byte[] TagSizeFromInteger(int n) {
            byte[] data = new byte[4];
            data[3] = (byte)(n & 0x7F);
            data[2] = (byte)((n >> 7) & 0x7F);
            data[1] = (byte)((n >> 7 * 2) & 0x7F);
            data[0] = (byte)((n >> 7 * 3) & 0x7F);
            return data;
        }
    }

    public class TID3V2Frame {
        public string frameId = "";
        public byte[] flags = null;
        public int bodyBytes = 0;
        public byte[] body = new byte[0];
        //
        public TTagVersion m_version = TTagVersion.V_2_2;
        int m_frameIDSize = 3;
        int m_dataSizeSize = 3;
        int m_flagsSize = 0;
        //
        public TID3V2Frame(TTagVersion value) {
            SetVersion(value);
            flags = new byte[m_flagsSize];
            for (int i = 0; i < m_flagsSize; i++) {
                flags[i] = 0;
            }
        }
        
        public TID3V2Frame Clone() {
            TID3V2Frame f = new TID3V2Frame(m_version);
            f.frameId = (string)frameId.Clone();
            f.bodyBytes = bodyBytes;
            f.body = (byte[])body.Clone();
            f.flags = (byte[])flags.Clone();
            return f;
        }

        public void SetVersion(TTagVersion value) {
            m_version = value;
            if (m_version == TTagVersion.V_2_2) {
                m_frameIDSize = 3;
                m_dataSizeSize = 3;
                m_flagsSize = 0;
            }
            else {
                m_frameIDSize = 4;
                m_dataSizeSize = 4;
                m_flagsSize = 2;
            }
        }
        
        public TTagString GetAsTagString() {
            TTagString s = new TTagString();
            int count = body.Length - 1;
            s.Set(body, 1, count, (TTagEncoding)body[0]);
            return s;
        }

        public string GetAsAsciiString() {
            if (body.Length > 1) {
                int count = body.Length - 1;
                if (body[1 + count - 1] == 0)
                    count -= 1;
                if (count > 0) {
                    ASCIIEncoding ae = new ASCIIEncoding();
                    return ae.GetString(body, 1, count);
                }
            }
            return "";
        }

        public TID3V2Comment GetAsID3V2Comment() {
            ASCIIEncoding ae = new ASCIIEncoding();
            TID3V2Comment c = new TID3V2Comment();
            c.language = ae.GetString(body, 1, 3);
            // LN^̃TCY
            int term = body[0] == (byte)TTagEncoding.ANSI ? 1 : 2;
            // fXNvṼoCg𓾂B
            int count = 0;
            for (int i = 4; ; i += term) {
                if (term == 2) {
                    if (body[i] == 0 && body[i + 1] == 0)
                        break;
                }
                else {
                    if (body[i] == 0)
                        break;
                }
                count += term;
            }

            c.description.Set(body, 4, count, (TTagEncoding)body[0]);
            c.text.Set(body, 4 + count + term, bodyBytes - (4 + count + term), (TTagEncoding)body[0]);

            return c;
        }

        /*
            v2.2
            Attached picture   "PIC"
            Frame size         $xx xx xx
            Text encoding      $xx
            Image format       $xx xx xx
            Picture type       $xx
            Description        <textstring> $00 (00)
            Picture data       <binary data>

            v2.3
            <Header for 'Attached picture', ID: "APIC">
            Text encoding      $xx
            MIME type          <text string> $00
            Picture type       $xx
            Description        <text string according to encoding> $00 (00)
            Picture data       <binary data>
         */
        public TTagPicture GetAsTagPicture() {
            ASCIIEncoding ae = new ASCIIEncoding();
            TTagPicture c = new TTagPicture();
            string imageFormat = "";
            int imageFormatSize = 0;
            if (m_version == TTagVersion.V_2_2) {
                imageFormatSize = 3;
                imageFormat = ae.GetString(body, 1, imageFormatSize);
            }
            else {
                int offset = 1;
                for (imageFormatSize = 0; body[offset + imageFormatSize] != 0; imageFormatSize++) {
                }
                imageFormat = ae.GetString(body, 1, imageFormatSize);
                if (imageFormat.Contains("png"))
                    imageFormat = "PNG";
                else if (imageFormat.Contains("jpeg"))
                    imageFormat = "JPG";
                imageFormatSize++;  // 0I[܂łf[^TCYB
            }
            byte pictureType = body[1 + imageFormatSize];
            // LN^̃TCY
            int term = body[0] == (byte)TTagEncoding.ANSI ? 1 : 2;
            // fXNvṼoCg𓾂B
            int count = 0;
            for (int i = 2 + imageFormatSize; ; i += term) {
                if (term == 2) {
                    if (body[i] == 0 && body[i + 1] == 0)
                        break;
                }
                else {
                    if (body[i] == 0)
                        break;
                }
                count += term;
            }
            TTagString description = new TTagString();
            description.Set(body, 2 + imageFormatSize, count, (TTagEncoding)body[0]);
            //description.DecodeString();
            byte[] pictureData = new byte[bodyBytes - (2 + imageFormatSize + count + term)];
            Array.Copy(body, 2 + imageFormatSize + count + term, pictureData, 0, pictureData.Length);
            c.Set(imageFormat, pictureType, description, pictureData);
            return c;
        }

        public bool Write(BinaryWriter bw) {
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] frameID = ae.GetBytes(this.frameId);
            bw.Write(frameID);
            bw.Write(SynchsafeInt.make(bodyBytes, m_dataSizeSize));
            //byte[] bodyBytes = new byte[m_dataSizeSize];
            //for (int i = 0; i < m_dataSizeSize; i++) {
            //    bodyBytes[i] = (byte)((m_bodyBytes >> (8 * (m_dataSizeSize - 1 - i))) & 0xFF);
            //}
            //bw.Write(bodyBytes);
            if (m_flagsSize > 0) {
                bw.Write(flags);
            }
            bw.Write(body);
            return true;
        }

        public bool MakeTextFrame(string inFrameId, TTagString ts, TTagVersion version) {
            this.frameId = inFrameId;
            bodyBytes = ts.original.Length + 1;
            body = new byte[bodyBytes];
            body[0] = (byte)ts.encoding;
            Array.Copy(ts.original, 0, body, 1, ts.original.Length);
            m_version = version;
            return true;
        }

        public bool WriteTextFrame(BinaryWriter bw, string frameID, TTagString ts, TTagVersion version) {
            if (MakeTextFrame(frameID, ts, version))
                return Write(bw);
            else
                return false;
        }

        public bool MakeAsciiFrame(string inFrameId, string s, TTagVersion version) {
            this.frameId = inFrameId;
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] bs = ae.GetBytes(s);
            bodyBytes = bs.Length + 1;
            body = new byte[bodyBytes];
            body[0] = (byte)TTagEncoding.ANSI;
            Array.Copy(bs, 0, body, 1, bs.Length);
            m_version = version;
            return true;
        }

        public bool WriteAsciiFrame(BinaryWriter bw, string inFrameId, string s, TTagVersion version) {
            if (MakeAsciiFrame(inFrameId, s, version))
                return Write(bw);
            else
                return false;
        }

        public bool MakeCommentFrame(string inFrameId, TTagString ts, TTagVersion version) {
            this.frameId = inFrameId;
            if (ts.original.Length == 0)
                return false;
            int descbytes = ts.encoding < TTagEncoding.UTF16 ? 1 : 4;
            bodyBytes = ts.original.Length + 1 /* encoding */ + 3 /* language */ + descbytes;
            body = new byte[bodyBytes];
            int p = 0;
            body[p] = (byte)ts.encoding;
            p++;
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] lang = ae.GetBytes("eng");
            Array.Copy(lang, 0, body, p, lang.Length);
            p += lang.Length;
            byte[] desc;
            if (ts.encoding < TTagEncoding.UTF16)
                desc = new byte[] { 0 };
            else
                desc = new byte[] { 0xFF, 0xFE, 0, 0 };
            Array.Copy(desc, 0, body, p, descbytes);
            p += descbytes;
            Array.Copy(ts.original, 0, body, p, ts.original.Length);
            return true;
        }

        public bool WriteCommentFrame(BinaryWriter bw, string inFrameId, TTagString ts, TTagVersion version) {
            if (MakeCommentFrame(inFrameId, ts, version))
                return Write(bw);
            else
                return false;
        }

        public bool MakePictureFrame(string inFrameId, TTagPicture tp, TTagVersion version) {
            this.frameId = inFrameId;
            if (tp.IsNull())
                return false;
            string imageFormat = tp.imageFormat;
            bodyBytes = 1;
            if (version == TTagVersion.V_2_2) {
                bodyBytes += 3;
            }
            else {
                if (imageFormat == "PNG")
                    imageFormat = "image/png";
                else if (imageFormat == "JPG")
                    imageFormat = "image/jpeg";
                else
                    return false;
                bodyBytes += imageFormat.Length + 1;
            }
            bodyBytes += 1;
            //tp.m_description.EncodeString(tp.m_description.m_encoding, -1);
            bodyBytes += tp.description.original.Length;
            bodyBytes += tp.description.encoding < TTagEncoding.UTF16 ? 1 : 2;    // 0I[B
            bodyBytes += tp.pictureData.Length;

            body = new byte[bodyBytes];
            int p = 0;
            body[p] = (byte)tp.description.encoding;
            p += 1;
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] f = ae.GetBytes(imageFormat);
            Array.Copy(f, 0, body, p, f.Length);
            p += f.Length;
            if (version > TTagVersion.V_2_2) {
                body[p] = 0;
                p += 1;
            }
            body[p] = (byte)tp.pictureType;
            p += 1;
            Array.Copy(tp.description.original, 0, body, p, tp.description.original.Length);
            p += tp.description.original.Length;
            if (tp.description.encoding < TTagEncoding.UTF16) {
                body[p] = 0;
                p += 1;
            }
            else {
                body[p] = 0;
                body[p + 1] = 0;
                p += 2;
            }
            Array.Copy(tp.pictureData, 0, body, p, tp.pictureData.Length);
            return true;
        }

        public bool WritePictureFrame(BinaryWriter bw, string frameID, TTagPicture tp, TTagVersion version) {
            if (MakePictureFrame(frameID, tp, version))
                return Write(bw);
            else
                return false;
        }

        public int Read(BinaryReader br) {
            ASCIIEncoding ae = new ASCIIEncoding();
            byte[] data = br.ReadBytes(m_frameIDSize);
            if (data[0] == 0)
                return m_frameIDSize;
            frameId = ae.GetString(data);
            bodyBytes = SynchsafeInt.parse(br.ReadBytes(m_dataSizeSize));
            //data = br.ReadBytes(m_dataSizeSize);
            //m_bodyBytes = 0;
            //for (int i = 0; i < m_dataSizeSize; i++) {
            //    m_bodyBytes += (data[i] << (8 * (m_dataSizeSize - 1 - i)));
            //}
            //m_bodyBytes = (data[0] << (8 * 2)) + (data[1] << (8 * 1)) + (data[2]);
            if (m_flagsSize > 0) {
                flags = br.ReadBytes(m_flagsSize);
            }
            if (bodyBytes > 0)
                body = br.ReadBytes(bodyBytes);
            return m_frameIDSize + m_dataSizeSize + m_flagsSize + bodyBytes;
        }
    }

}
