using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Windows.Forms;
using System.Threading;

namespace ID3TagEditLite {
    public class TPluginItem {
        public string filePath;
        public TPluginManager mgr;

        public bool Execute(TTagItem tag) {
            TPluginInfo info = mgr.pluginInfos[mgr.GetExtensionForKey(filePath)];
            if (info == null)
                return false;
            string exe = info.m_exe;
            if (info.m_exe.Length == 0)
                exe = filePath;
            string param = info.m_param;
            if (info.m_exe.Length > 0) {
                if (param.Length > 0)
                    param += " ";
                param += "\"" + filePath + "\"";
            }

            List<GetTagStringHandler> getCallbackTable = tag.record.getTagStringCallbackHandlerTable;
            List<SetTagStringHandler> setCallbackTable = tag.record.setTagStringCallbackHandlerTable;

            Encoding enc;
            if (mgr.asis.Contains(info.m_encoding.ToLower()))
                enc = Encoding.GetEncoding(Program.thePref.ansiCodePage);
            else
                enc = new UTF8Encoding(false, false);

            Process proc = new Process();
            proc.StartInfo.FileName = exe;
            proc.StartInfo.Arguments = param;
            proc.StartInfo.UseShellExecute = false;
            proc.StartInfo.RedirectStandardInput = true;
            proc.StartInfo.RedirectStandardOutput = true;
            //proc.StartInfo.WindowStyle = ProcessWindowStyle.Minimized;
            proc.StartInfo.CreateNoWindow = true;
            proc.StartInfo.StandardOutputEncoding = enc;
            proc.Start();

            StreamWriter sw = new StreamWriter(proc.StandardInput.BaseStream, enc);
            StreamReader sr = proc.StandardOutput;
            string path = tag.filePath;
            path = path.Replace('\\', '/');
            sw.WriteLine(path);
            foreach (GetTagStringHandler getCallback in getCallbackTable) {
                sw.WriteLine(getCallback());
            }
            sw.Close();
            path = sr.ReadLine();
            foreach (SetTagStringHandler setCallback in setCallbackTable) {
                string s = sr.ReadLine();
                if (s == null)
                    break;
                setCallback(s);
            }
            sr.Close();
            proc.WaitForExit();
            proc.Close();
            return true;
        }

        bool IsAsIs(string s) {
            return mgr.asis.Contains(s.ToLower());
        }

        public string GetNameForMenu() {
            return Path.GetFileNameWithoutExtension(filePath);
        }

        public void OnClick(Object sender, EventArgs e) {
            mgr.ExecutePlugin(this);
            //List<TTagItem> tags = mgr.GetSelectedTagItems();
            //if (tags == null || tags.Count == 0)
            //    return;
            //mgr.SetUndo(tags);
            //int percent = 0;
            //mgr.Progress(0);
            //int i = 0;
            //foreach (TTagItem t in tags) {
            //    i++;
            //    percent = 100 * i / tags.Count;
            //    mgr.Progress(percent);
            //    Execute(t);
            //}
            //mgr.Progress(0);
            //mgr.Redraw(tags);
        }
    }

    public class TPluginInfo {
        public string m_ext = "";
        public string m_exe = "";
        public string m_param = "";
        public string m_encoding = "";
    }

    public class TPluginManager {
        public SortedDictionary<string, TPluginItem> plugins = new SortedDictionary<string, TPluginItem>();
        public SortedDictionary<string, TPluginInfo> pluginInfos = new SortedDictionary<string, TPluginInfo>();
        public List<string> asis = new List<string>();
        public string redoName = "";
        public ProgressHandler ProgressCallback = null;
        public RedrawHandler RedrawCallback = null;
        public GetSelectedTagItemsHandler GetSelectedTagItemsCallback = null;
        public SetUndoHandler SetUndoCallback = null;
        public WorkingHandler WorkingCallback = null;

        Object cueLock = new Object();
        Thread th = null;
        bool working = false;
        bool stopRequest = false;

        public void Setup() {
            string moduleFilePath = Process.GetCurrentProcess().MainModule.FileName;
            string dirPath = Path.GetDirectoryName(moduleFilePath);
            string pluginInfoPath = Path.Combine(dirPath, "Plugin.info");
            LoadPluginInfoFile(pluginInfoPath);
            string pluginDirPath = Path.Combine(dirPath, "Plugin");
            ListupPluginDirectory(pluginDirPath);
            asis.Add("asis");
            asis.Add("as_is");
            asis.Add("as is");
            asis.Add("as-is");
            asis.Add("ansi");
            asis.Add("ascii");
            asis.Add("jis");
            asis.Add("sjis");
            asis.Add("shift jis");
            asis.Add("shiftjis");
            asis.Add("shift_jis");
            asis.Add("shift-jis");
            asis.Add("windows31j");
            asis.Add("nihongo");
            asis.Add("kanji");
        }

        public bool LoadPluginInfoFile(string filePath) {
            if (File.Exists(filePath) == false)
                return false;
            FileStream f = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            if (f == null)
                return false;
            StreamReader sr = new StreamReader(f);
            char[] dlmtTab = new char[] { '\t' };
            char[] dlmtComma = new char[] { ',', ' ' };
            while (true) {
                string s = sr.ReadLine();
                if (s == null)
                    break;
                if (s.Length == 0)
                    continue;
                string[] ss = s.Split(dlmtTab);
                if (ss.Length == 0)
                    continue;
                string[] exts = ss[0].Split(dlmtComma, StringSplitOptions.RemoveEmptyEntries);
                List<TPluginInfo> infos = new List<TPluginInfo>();
                foreach (string ext in exts) {
                    TPluginInfo info = new TPluginInfo();
                    info.m_ext = ext;
                    infos.Add(info);
                }
                if (ss.Length > 1) {
                    foreach (TPluginInfo info in infos)
                        info.m_exe = ss[1];
                }
                if (ss.Length > 2) {
                    foreach (TPluginInfo info in infos)
                        info.m_param = ss[2];
                }
                if (ss.Length > 3) {
                    foreach (TPluginInfo info in infos)
                        info.m_encoding = ss[3];
                }
                foreach (TPluginInfo info in infos)
                    pluginInfos.Add(info.m_ext, info);
            }
            sr.Close();
            f.Close();
            return true;
        }

        public string GetExtensionForKey(string path) {
            string ext = Path.GetExtension(path);
            if (ext.Length == 0)
                ext = ".";
            return ext;
        }

        bool ListupPluginDirectory(string dirPath) {
            if (Directory.Exists(dirPath) == false)
                return false;
            string[] files = Directory.GetFiles(dirPath);
            foreach (string path in files) {
                string ext = GetExtensionForKey(path);
                if (pluginInfos.ContainsKey(ext) == false)
                    continue;
                TPluginItem p = new TPluginItem();
                p.filePath = path;
                p.mgr = this;
                string key = p.GetNameForMenu();
                plugins.Add(key, p);
            }
            return true;
        }

        TPluginItem GetPluginItemByMenuName(string name) {
            return plugins[name];
        }

        public void Redo() {
            if (redoName.Length == 0) {
                return;
            }
            if (!plugins.ContainsKey(redoName)) {
                return;
            }
            plugins[redoName].OnClick(null, null);
        }

        public void Redraw(List<TTagItem> tags) {
            if (IsContinue() && RedrawCallback != null) {
                RedrawCallback(tags);
            }
        }

        public void Progress(int percent) {
            if (IsContinue() && ProgressCallback != null) {
                ProgressCallback(percent);
            }
        }

        //public List<TTagItem> GetSelectedTagItems() {
        //    if (GetSelectedTagItemsCallback != null) {
        //        return GetSelectedTagItemsCallback();
        //    }
        //    return new List<TTagItem>();
        //}

        //public void SetUndo(List<TTagItem> tags) {
        //    if (SetUndoCallback != null) {
        //        SetUndoCallback(tags);
        //    }
        //}

        public bool IsWorking() {
            lock (cueLock) {
                return working;
            }
        }

        public void Stop() {
            lock (cueLock) {
                stopRequest = true;
            }
            if (th != null) {
                th.Join();
            }
            if (ProgressCallback != null) {
                ProgressCallback(0);
            }
            if (RedrawCallback != null) {
                RedrawCallback(tags);
            }
        }

        private bool IsContinue() {
            lock (cueLock) {
                return !stopRequest;
            }
        }

        TPluginItem currentPlugin = null;
        public List<TTagItem> tags = new List<TTagItem>();

        void ExecutePluginThread() {
            int percent = 0;
            Progress(0);
            int i = 0;
            foreach (TTagItem t in tags) {
                if (!IsContinue()) {
                    break;
                }
                i++;
                percent = 100 * i / tags.Count;
                Progress(percent);
                currentPlugin.Execute(t);
            }
            Progress(0);
            Redraw(tags);
            working = false;
        }

        public void ExecutePlugin(TPluginItem plugin) {
            if (WorkingCallback()) {
                return;
            }
            tags = GetSelectedTagItemsCallback();
            if (tags.Count == 0) {
                return;
            }
            SetUndoCallback(tags);
            currentPlugin = plugin;
            redoName = currentPlugin.GetNameForMenu();
            stopRequest = false;
            th = new Thread(new ThreadStart(ExecutePluginThread));
            working = true;
            th.Start();
        }
    }
}
