using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Xml.Serialization;
using System.IO;
using System.Globalization;

namespace ID3TagEditLite {
    public class Preferences {
        const int CURRENT_PREF_VERSION = 0x100;

        public int version = CURRENT_PREF_VERSION;
        public int ansiCodePage = CultureInfo.CurrentCulture.TextInfo.ANSICodePage;
        public Rectangle userBounds;
        public Point editDlgLocation;
        public Point genreDlgLocation;
        //public Point location = new Point(0, 0);
        //public int windowHeight = 0;
        //public int windowWidth = 0;
        public string loadDir = "";
        public int defaultSaveEnc = (int)TTagEncoding.UTF16;
        public int defaultVersion = (int)TTagVersion.V_2_2;
        public int exportIndex = 0;
        public int importIndex = 0;
        public String exportDir = "";
        public String importDir = "";
        public int columnWidthModify = 20;
        public int columnWidthTrack = 30;
        public int columnWidthFileName = 100;
        public int columnWidthTitle = 100;
        public int columnWidthArtist = 100;
        public int columnWidthAlbum = 100;
        public int columnWidthYear = 40;
        public int columnWidthComment = 20;
        public int columnWidthGenre = 40;
        public int columnWidthVersion = 30;
        public int columnWidthSourceEncode = 40;
        public int columnWidthTargetEncode = 40;
        public bool nothingIsAll = true;
        public bool autoAppendCoverImage = false;
        public bool copyOnlyTitleAndArtist = false;
        public string fontFamily = System.Drawing.SystemFonts.MessageBoxFont.FontFamily.Name;
        public float fontSize = System.Drawing.SystemFonts.MessageBoxFont.Size;
        public bool applyFontToDialog = true;

        /// <summary>
        /// f[^̃Rs[
        /// </summary>
        /// <param name="s"></param>
        public void CopyFrom(Preferences s) {
            version = s.version;
            ansiCodePage = s.ansiCodePage;
            userBounds = s.userBounds;
            editDlgLocation = s.editDlgLocation;
            genreDlgLocation = s.genreDlgLocation;
            //location = s.location;
            //windowHeight = s.windowHeight;
            //windowWidth = s.windowWidth;
            loadDir = s.loadDir;
            defaultSaveEnc = s.defaultSaveEnc;
            defaultVersion = s.defaultVersion;
            exportIndex = s.exportIndex;
            importIndex = s.importIndex;
            exportDir = s.exportDir;
            importDir = s.importDir;
            columnWidthModify = s.columnWidthModify;
            columnWidthTrack = s.columnWidthTrack;
            columnWidthFileName = s.columnWidthFileName;
            columnWidthTitle = s.columnWidthTitle;
            columnWidthArtist = s.columnWidthArtist;
            columnWidthAlbum = s.columnWidthAlbum;
            columnWidthYear = s.columnWidthYear;
            columnWidthComment = s.columnWidthComment;
            columnWidthGenre = s.columnWidthGenre;
            columnWidthVersion = s.columnWidthVersion;
            columnWidthSourceEncode = s.columnWidthSourceEncode;
            columnWidthTargetEncode = s.columnWidthTargetEncode;
            nothingIsAll = s.nothingIsAll;
            autoAppendCoverImage = s.autoAppendCoverImage;
            copyOnlyTitleAndArtist = s.copyOnlyTitleAndArtist;
            fontFamily = s.fontFamily;
            fontSize = s.fontSize;
            applyFontToDialog = s.applyFontToDialog;
        }

        static string GetSettingDirPath() {
            return Path.Combine(System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OSTRA");
        }

        static string GetSettingFilePath() {
            return Path.Combine(GetSettingDirPath(), "ID3TagEditLite_1_0.pref");
        }

        /// <summary>
        /// ݒ̓ǂݍ
        /// </summary>
        /// <returns></returns>
        public bool Beginning() {
            bool result = false;
            Preferences temp = new Preferences();
            XmlSerializer serializer = new XmlSerializer(typeof(Preferences));
            FileStream fs = null;
            try {
                fs = new FileStream(GetSettingFilePath(), FileMode.Open);
            }
            catch {
                fs = null;
            }
            if (fs != null) {
                try {
                    temp = (Preferences)serializer.Deserialize(fs);
                }
                catch {
                }
                fs.Close();
                if (temp.version != 0) {
                    CopyFrom(temp);
                    version = CURRENT_PREF_VERSION;
                    result = true;
                }
            }

            return result;
        }

        /// <summary>
        /// ݒ̏o
        /// </summary>
        /// <returns></returns>
        public bool Ending() {
            bool result = false;
            string dir = GetSettingDirPath();
            if (Directory.Exists(dir) == false) {
                Directory.CreateDirectory(dir);
            }
            XmlSerializer serializer = new XmlSerializer(typeof(Preferences));
            FileStream fs = new FileStream(GetSettingFilePath(), FileMode.Create);
            if (fs != null) {
                serializer.Serialize(fs, this);
                fs.Close();
                result = true;
            }
            return result;
        }

    }
}
