using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Collections;
using System.Globalization;
using System.Windows.Forms;
using Uguu.IO.Csv;
using System.Threading;

namespace ID3TagEditLite {

    public delegate string GetTagStringHandler();
    public delegate void SetTagStringHandler(string s);

    /// <summary>
    /// GR[fBO̎ށB
    /// </summary>
    public enum TTagEncoding {
//        LATIN1 = -1,    // tO
        ANSI = 0,      // ID3 Tag V2
        UTF16 = 1
    }

    public enum TTagVersion {
        V_1_0 = 0,      // tO
        V_1_1 = 1,
        V_2_2 = 2,      // ID2 Tag V2
        V_2_3 = 3,
        V_2_4 = 4
    }

    public enum TTagPictureType {
        OTHER = 0,
        ICON_32 = 1,
        ICON = 2,
        COVER_FRONT = 0x3
    }

    public class SynchsafeInt {
        public static int parse(byte[] data) {
            int value = 0;
            for (int i = 0; i < data.Length; i++) {
                value += (data[i] << (8 * (data.Length - 1 - i)));
            }
            return value;
        }

        public static byte[] make(int value, int sz) {
            byte[] data = new byte[sz];
            for (int i = 0; i < sz; i++) {
                data[i] = (byte)((value >> (8 * (sz - 1 - i))) & 0xFF);
            }
            return data;
        }
    }

    // œƗt@CɂH
    public class TTagPicture {
        public string imageFormat = "";
        public byte pictureType = (byte)TTagPictureType.OTHER;
        public TTagString description = new TTagString();
        public byte[] pictureData = null;

        public void Init() {
            imageFormat = "";
            pictureType = (byte)TTagPictureType.OTHER;
            description = new TTagString();
            pictureData = null;
        }

        public void Set(string imageFormat, byte pictureType, TTagString description, byte[] pictureData) {
            this.imageFormat = imageFormat;
            this.pictureType = pictureType;
            this.description = description;
            this.pictureData = pictureData;
        }

        public TTagPicture Clone() {
            TTagPicture c = new TTagPicture();
            c.imageFormat = (string)imageFormat.Clone();
            c.pictureType = pictureType;
            c.description = description.Clone();
            c.imageFormat = (string)imageFormat.Clone();
            // 摜f[^͕̂̂QƂg܂킷ƂɂĂ݂B
            c.pictureData = pictureData;
            //c.m_pictureData = (byte[])m_pictureData.Clone();
            return c;
        }

        public bool IsNull() {
            return pictureData == null || pictureData.Length == 0;
        }

        string FindCoverFileName(string mp3FilePath) {
            List<string> pathList = new List<string>();
            pathList.Add(Path.ChangeExtension(mp3FilePath, ".png"));
            pathList.Add(Path.ChangeExtension(mp3FilePath, ".jpg"));
            pathList.Add(Path.Combine(Path.GetDirectoryName(mp3FilePath), "folder.png"));
            pathList.Add(Path.Combine(Path.GetDirectoryName(mp3FilePath), "folder.jpg"));
            pathList.Add(Path.Combine(Path.GetDirectoryName(mp3FilePath), "cover.png"));
            pathList.Add(Path.Combine(Path.GetDirectoryName(mp3FilePath), "cover.jpg"));
            foreach (string path in pathList) {
                try {
                    FileInfo info = new FileInfo(path);
                    if (info.Exists && info.Length > 0)
                        return path;
                }
                catch {
                }
            }
            return "";
        }

        public bool LoadCoverFile(string mp3FilePath) {
            string path = FindCoverFileName(mp3FilePath);
            if (path.Length == 0)
                return false;
            FileInfo info = new FileInfo(path);
            FileStream fs = null;
            BinaryReader br = null;
            try {
                fs = new FileStream(path, FileMode.Open, FileAccess.Read);
                br = new BinaryReader(fs);
                pictureData = br.ReadBytes((int)info.Length);
            }
            catch {
                return false;
            }
            br.Close();
            fs.Close();
            if (path.EndsWith(".png"))
                imageFormat = "PNG";
            else
                imageFormat = "JPG";
            pictureType = (byte)TTagPictureType.OTHER;
            return true;
        }
    }

    /// <summary>
    /// Undo/Redo₷邽߂̃^Of[^̃NXB
    /// </summary>
    public class TTagRecord {
        public bool bModify = false;
        public string fileName = "";
        public TTagString title = new TTagString();
        public TTagString artist = new TTagString();
        public TTagString album = new TTagString();
        public TTagString comment = new TTagString();
        public string year = "";
        public int genre = 0;
        public TTagVersion version = TTagVersion.V_2_2;
        public int track = 0;
        public int trackMax = 0;
        public TTagEncoding loadEnc = (TTagEncoding)Program.thePref.defaultSaveEnc;
        public TTagEncoding saveEnc = (TTagEncoding)Program.thePref.defaultSaveEnc;
        public TTagPicture picture = new TTagPicture();
        // fXNvVRgB
        public List<TID3V2Frame> systemComment = new List<TID3V2Frame>();
        // 摜B
        public List<TID3V2Frame> otherPicture = new List<TID3V2Frame>();
        // ̑
        public List<TID3V2Frame> other = new List<TID3V2Frame>();

        public List<GetTagStringHandler> getTagStringCallbackHandlerTable;
        public List<SetTagStringHandler> setTagStringCallbackHandlerTable;

        public TTagRecord() {
            getTagStringCallbackHandlerTable = CreateGetStringCallbackTable();
            setTagStringCallbackHandlerTable = CreateSetStringCallbackTable();
        }

        public TTagRecord Clone() {
            TTagRecord r = new TTagRecord();
            r.fileName = (string)fileName.Clone();
            r.title = title.Clone();
            r.artist = artist.Clone();
            r.album = album.Clone();
            r.comment = comment.Clone();
            r.year = (string)year.Clone();
            r.genre = genre;
            r.version = version;
            r.track = track;
            r.trackMax = trackMax;
            r.loadEnc = loadEnc;
            r.saveEnc = saveEnc;
            r.picture = picture.Clone();
            foreach (TID3V2Frame c in systemComment)
                r.systemComment.Add(c.Clone());
            foreach (TID3V2Frame f in other)
                r.other.Add(f.Clone());
            foreach (TID3V2Frame f in otherPicture)
                r.otherPicture.Add(f.Clone());
            r.bModify = bModify;
            return r;
        }

        public bool DecodeStrings() {
            Boolean mod = false;
            if (title.DecodeString(/*m_loadEnc*/))
                mod = true;
            if (artist.DecodeString(/*m_loadEnc*/))
                mod = true;
            if (album.DecodeString(/*m_loadEnc*/))
                mod = true;
            if (comment.DecodeString(/*m_loadEnc*/))
                mod = true;
            if (picture.description.DecodeString())
                mod = true;
            return mod;
        }

        public bool EncodeStrings() {
            if (version >= TTagVersion.V_2_2) {
                title.EncodeString(saveEnc, -1);
                artist.EncodeString(saveEnc, -1);
                album.EncodeString(saveEnc, -1);
                comment.EncodeString(saveEnc, -1);
                picture.description.EncodeString(picture.description.encoding, -1);
            }
            else {
                title.EncodeString(saveEnc, 30);
                artist.EncodeString(saveEnc, 30);
                album.EncodeString(saveEnc, 30);
                comment.EncodeString(saveEnc, version == TTagVersion.V_1_1 ? 28 : 30);
            }
            return true;
        }

        public string GetModifyString() {
            return bModify ? "+" : "-";
        }

        public string GetFileName() {
            return fileName;
        }

        public string GetTitle() {
            return title.str;
        }

        public string GetArtist() {
            return artist.str;
        }
        
        public string GetAlbum() {
            return album.str;
        }

        public string GetYear() {
            return year;
        }

        public string GetComment() {
            return comment.str;
        }

        public string GetVersionString() {
            if ((int)version >= 0 && (int)version < 5)
                return GetVersionNames()[(int)version];
            else
                return "-";
        }

        public string GetGenreNumberString() {
            return genre.ToString();
        }

        public string GetGenreString() {
            string g = "";
            if (Program.theGenre.GetGenreNameByGenreNumber(ref g, genre) && g.Length > 0)
                return g;
            else
                return GetGenreNumberString();
        }

        public string GetGenreExportString() {
            return "(" + genre.ToString() + ")";
        }

        public string GetTrackAndTrackMaxString() {
            if (trackMax > 0)
                return GetTrackString() + "/" + GetTrackMaxString();
            else
                return GetTrackString();
        }

        public string GetTrackString() {
            if (track > 0)
                return track.ToString();
            else
                return "";
        }

        public string GetTrackMaxString() {
            if (trackMax > 0)
                return trackMax.ToString();
            else
                return "";
        }

        public static string GetEncodingString(TTagEncoding coding) {
//            if (coding == TTagEncoding.LATIN1) {
//                return "LATIN-1";
//            }
            Encoding enc;
            if (coding == TTagEncoding.UTF16) {
                enc = Encoding.Unicode;
            }
            else {
                enc = Encoding.GetEncoding(Program.thePref.ansiCodePage);
            }
            return enc.WebName.ToUpper();
        }

        public static string[] GetEncodingNames() {
            string[] ss = new string[2];
            ss[0] = GetEncodingString(TTagEncoding.ANSI);
            ss[1] = GetEncodingString(TTagEncoding.UTF16);
            return ss;
        }

        public static string[] GetVersionNames() {
            return new string[] { "v1.0", "v1.1", "v2.2", "v2.3", "v2.4" };
        }

        public static string CorrectYear(string year) {
            int y = 0;
            if (TTools.StringToInteger(ref y, year) && y > 0)
                return y.ToString();
            return "";
        }

        public void SetPropertyOfGenre(string value) {
            char[] dlmt = { '(', ')' };
            string[] ss = value.Split(dlmt, StringSplitOptions.RemoveEmptyEntries);
            if (ss.Length > 0)
                SetIntegerProperty(ref genre, ss[0]);
        }

        public void SetPropertyOfTrack(string value) {
            SetIntegerProperty(ref track, value);
        }

        public void SetPropertyOfTrackMax(string value) {
            SetIntegerProperty(ref trackMax, value);
        }

        public void SetPropertyOfYear(string value) {
            string ys = CorrectYear(value);
            SetStringProperty(ref year, ys);
        }

        public void SetPropertyOfFileName(string value) {
            SetStringProperty(ref fileName, value);
        }

        public void SetPropertyOfTitle(string value) {
            SetTagStringProperty(ref title, value);
        }

        public void SetPropertyOfArtist(string value) {
            SetTagStringProperty(ref artist, value);
        }

        public void SetPropertyOfAlbum(string value) {
            SetTagStringProperty(ref album, value);
        }

        public void SetPropertyOfComment(string value) {
            SetTagStringProperty(ref comment, value);
        }

        public void SetPropertyOfVersion(TTagVersion value) {
            if (version != value) {
                version = value;
                bModify = true;
            }
        }

        public void SetPropertyOfSaveEnc(TTagEncoding value) {
            if (saveEnc != value) {
                saveEnc = value;
                bModify = true;
            }
        }

        public void SetPropertyOfLoadEnc(TTagEncoding value) {
            if (loadEnc != value) {
                loadEnc = value;
                bModify = true;
            }
        }

        public void SetStringProperty(ref string var, string value) {
            if (var.Equals(value) == false) {
                var = value;
                bModify = true;
            }
        }

        public void SetTagStringProperty(ref TTagString var, string value) {
            if (var.str.Equals(value) == false) {
                var.str = value;
                bModify = true;
            }
        }

        public void SetIntegerProperty(ref int var, string value) {
            int n = 0;
            if (value.Length == 0) {
                var = n;
                bModify = true;
            }
            else if (TTools.StringToInteger(ref n, value)) {
                if (n != var) {
                    var = n;
                    bModify = true;
                }
            }
        }

        List<GetTagStringHandler> CreateGetStringCallbackTable() {
            List<GetTagStringHandler> table = new List<GetTagStringHandler>();
            table.Add(new GetTagStringHandler(GetTrackString));
            table.Add(new GetTagStringHandler(GetFileName));
            table.Add(new GetTagStringHandler(GetTitle));
            table.Add(new GetTagStringHandler(GetAlbum));
            table.Add(new GetTagStringHandler(GetArtist));
            table.Add(new GetTagStringHandler(GetYear));
            table.Add(new GetTagStringHandler(GetComment));
            table.Add(new GetTagStringHandler(GetGenreExportString));
            return table;
        }

        List<SetTagStringHandler> CreateSetStringCallbackTable() {
            List<SetTagStringHandler> table = new List<SetTagStringHandler>();
            table.Add(new SetTagStringHandler(SetPropertyOfTrack));
            table.Add(new SetTagStringHandler(SetPropertyOfFileName));
            table.Add(new SetTagStringHandler(SetPropertyOfTitle));
            table.Add(new SetTagStringHandler(SetPropertyOfAlbum));
            table.Add(new SetTagStringHandler(SetPropertyOfArtist));
            table.Add(new SetTagStringHandler(SetPropertyOfYear));
            table.Add(new SetTagStringHandler(SetPropertyOfComment));
            table.Add(new SetTagStringHandler(SetPropertyOfGenre));
            return table;
        }

        public string GetTSVLine() {
            List<GetTagStringHandler> callbackTable = getTagStringCallbackHandlerTable;
            bool bFirst = true;
            string result = "";
            foreach (GetTagStringHandler callback in callbackTable) {
                string s = callback();
                if (bFirst == false)
                    result += "\t";
                result += s;
                bFirst = false;
            }
            return result;
        }

        public void SetTSVLine(string s) {
            List<SetTagStringHandler> callbackTable = setTagStringCallbackHandlerTable;
            char[] dlmt = new char[] { '\t' };
            string[] data = s.Split(dlmt);
            for (int i = 0; i < callbackTable.Count && i < data.Length; i++) {
                callbackTable[i](data[i]);
            }
        }

        public string GetCSVLine() {
            UnicodeEncoding enc = new UnicodeEncoding(false, false);
            MemoryStream ms = new MemoryStream();
            StreamWriter sw = new StreamWriter(ms, enc);
            CsvWriter cw = new CsvWriter(sw);
            List<GetTagStringHandler> getCallbackTable = getTagStringCallbackHandlerTable;
            foreach (GetTagStringHandler getCallback in getCallbackTable) {
                cw.Write(getCallback());
            }
            cw.Close();
            sw.Close();
            ms.Close();
            byte[] data = ms.ToArray();
            string s = enc.GetString(data);
            return s;
        }

        public void SetCSVLine(string s) {
            UnicodeEncoding enc = new UnicodeEncoding(false, false);
            MemoryStream ms = new MemoryStream(enc.GetBytes(s));
            StreamReader sr = new StreamReader(ms, enc);
            CsvReader cr = new CsvReader(sr);
            List<SetTagStringHandler> setCallbackTable = setTagStringCallbackHandlerTable;
            if (cr.Read() == false)
                return;
            for (int i = 0; i < cr.FieldCount && i < setCallbackTable.Count; i++) {
                setCallbackTable[i](cr[i]);
            }
            cr.Close();
            sr.Close();
            ms.Close();
        }    
    }

    /// <summary>
    /// 1MP3t@CɊ܂܂^OŜ̃NXB
    /// </summary>
    public class TTagItem {
        public TTagRecord record = new TTagRecord();
        public string filePath = "";
        public long soundStartPos = 0;
        public long soundSize = 0;
        public TTagRecord undoRecord = null;
        public TTagRecord redoRecord = null;
        public string tempFilePath = "";
        public TTagManager container = null;

        public TTagItem(TTagManager container) {
            this.container = container;
        }
        
        /// <summary>
        /// t@CpXw肵ID3^OǂݍށB
        /// </summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        public bool LoadTag(string filePath) {
            this.filePath = filePath;
            record.fileName = Path.GetFileName(this.filePath);
            soundStartPos = 0;
            try {
                FileInfo info = new FileInfo(this.filePath);
                soundSize = info.Length;
            }
            catch {
                return false;
            }

            record.bModify = false;

            LoadV1Tag();
            LoadV2Tag();

            if (Program.thePref.autoAppendCoverImage) {
                // Jo[摜荞݁B
                if (record.picture.IsNull()) {
                    if (record.picture.LoadCoverFile(this.filePath)) {
                        record.picture.description.DecodeString();
                        record.bModify = true;
                    }
                }
            }

            return true;
        }

        bool LoadV1Tag() {
            FileInfo info = new FileInfo(filePath);
            if (info.Length < TID3V1Tag.TAG_SIZE)
                return false;
            FileStream fs = null;
            BinaryReader br = null;
            try {
                fs = new FileStream(filePath, FileMode.Open, FileAccess.Read);
                br = new BinaryReader(fs);
            }
            catch {
                return false;
            }
            TID3V1Tag tag = new TID3V1Tag();
            br.BaseStream.Seek(info.Length - TID3V1Tag.TAG_SIZE, SeekOrigin.Begin);
            bool result = tag.Read(br);
            br.Close();
            fs.Close();
            if (result == false)
                return false;

            record.version = tag.zero == 0 ? TTagVersion.V_1_1 : TTagVersion.V_1_0;

            ASCIIEncoding ae = new ASCIIEncoding();
            record.title.Set(tag.title, 0, tag.title.Length, TTagEncoding.ANSI);
            record.artist.Set(tag.artist, 0, tag.artist.Length, TTagEncoding.ANSI);
            record.album.Set(tag.album, 0, tag.album.Length, TTagEncoding.ANSI);
            record.year = TTagRecord.CorrectYear(ae.GetString(tag.year));
            record.comment.Set(tag.comment, 0, tag.comment.Length, TTagEncoding.ANSI);
            record.track = tag.track;
            record.genre = tag.genre;

            soundSize = info.Length - TID3V1Tag.TAG_SIZE;
            record.loadEnc = TTagEncoding.ANSI;
            record.saveEnc = record.loadEnc;
            record.DecodeStrings();

            return true;
        }
        
        const int FLAGS_UNSYNCHRONISATION = 0x80;
        const int FLAGS_EXTENDED_HEADER = 0x40;

        bool LoadV2Tag() {
            FileInfo info = new FileInfo(filePath);
            if (info.Length < TID3V2Head.HEAD_SIZE)
                return false;
            FileStream fs = null;
            BinaryReader br = null;
            try {
                fs = new FileStream(filePath, FileMode.Open, FileAccess.Read);
                br = new BinaryReader(fs);
            }
            catch {
                return false;
            }
            TID3V2Head head = new TID3V2Head();
            bool result = head.Read(br);
            if (result) {
                soundStartPos = TID3V2Head.HEAD_SIZE + head.size;
                soundSize -= soundStartPos;
                record.version = (TTagVersion)head.major;
                if (record.version > TTagVersion.V_2_2 && (head.flags & FLAGS_UNSYNCHRONISATION) != 0) {
                    // bit 7 񓯊ɂ͔ΉB
                    return false;
                }
                if (record.version > TTagVersion.V_2_2 && (head.flags & FLAGS_EXTENDED_HEADER) != 0) {
                    // bit 6 gwb_͓ǂݎ̂ĂB
                    br.ReadBytes(SynchsafeInt.parse(br.ReadBytes(4)));
                }
                result = LoadV2TagSub(br, head.size);
            }
            br.Close();
            fs.Close();
            if (result == false)
                return false;

            // łUNICODEƐݒ肵ĂȂ΁AŜUNICODEňB
            // Ŝߔ > 0 ƂĂB
            if (record.title.encoding > TTagEncoding.ANSI ||
                    record.artist.encoding > TTagEncoding.ANSI ||
                    record.album.encoding > TTagEncoding.ANSI ||
                    record.comment.encoding > TTagEncoding.ANSI) {
                record.loadEnc = TTagEncoding.UTF16;
            }
            else {
                /* Ȃ`FbNKv */
                record.loadEnc = TTagEncoding.ANSI;
                if (record.loadEnc == TTagEncoding.ANSI)
                    record.loadEnc = record.title.CheckString();
                if (record.loadEnc == TTagEncoding.ANSI)
                    record.loadEnc = record.artist.CheckString();
                if (record.loadEnc == TTagEncoding.ANSI)
                    record.loadEnc = record.album.CheckString();
                if (record.loadEnc == TTagEncoding.ANSI)
                    record.loadEnc = record.comment.CheckString();
            }

            //if (Program.thePref.autoAppendCoverImage) {
            //    // Jo[摜荞݁B
            //    if (m_record.m_picture.IsNull()) {
            //        if (m_record.m_picture.LoadCoverFile(m_filePath)) {
            //            m_record.m_modify = true;
            //        }
            //    }
            //}

            record.saveEnc = record.loadEnc;
            record.DecodeStrings();

            return true;
        }

        bool LoadV2TagSub(BinaryReader br, int tagSize) {
            int pos = TID3V2Head.HEAD_SIZE;
            while (pos < tagSize) {
                TID3V2Frame frame = new TID3V2Frame(record.version);
                pos += frame.Read(br);
                if (frame.frameId.Length == 0) {
                    break;
                }
                else if (frame.frameId.Equals("TT2") || frame.frameId.Equals("TIT2")) {
                    record.title = frame.GetAsTagString();
                }
                else if (frame.frameId.Equals("TP1") || frame.frameId.Equals("TPE1")) {
                    record.artist = frame.GetAsTagString();
                }
                else if (frame.frameId.Equals("TAL") || frame.frameId.Equals("TALB")) {
                    record.album = frame.GetAsTagString();
                }
                else if (frame.frameId.Equals("TYE") || frame.frameId.Equals("TYER")) {
                    record.year = frame.GetAsAsciiString();
                }
                else if (frame.frameId.Equals("COM") || frame.frameId.Equals("COMM")) {
                    TID3V2Comment c = frame.GetAsID3V2Comment();
                    c.DecodeString(c.text.encoding);
                    if (c.description.str.Length > 0)
                        record.systemComment.Add(frame);
                    else
                        record.comment = c.text;
                }
                else if (frame.frameId.Equals("PIC") || frame.frameId.Equals("APIC")) {
                    TTagPicture p = frame.GetAsTagPicture();
                    if (record.picture.IsNull() && (p.pictureType == (byte)TTagPictureType.OTHER || p.pictureType == (byte)TTagPictureType.COVER_FRONT))
                        record.picture = p;
                    else
                        record.otherPicture.Add(frame);
                }
                else if (frame.frameId.Equals("TCO") || frame.frameId.Equals("TCON")) {
                    string[] t = frame.GetAsTagString().GetDecodeString().Split(new char[] { '(', ')' }, StringSplitOptions.RemoveEmptyEntries);
                    //string[] t = frame.GetAsAsciiString().Split(new char[] { '(', ')' }, StringSplitOptions.RemoveEmptyEntries);
                    if (t.Length > 0) {
                        if (TTools.StringToInteger(ref record.genre, t[0]) == false) {
                            string[] u = t[0].Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
                            foreach (string s in u) {
                                if (Program.theGenre.SearchGenreNumberByGenreName(ref record.genre, s))
                                    break;
                            }
                        }
                    }
                }
                else if (frame.frameId.Equals("TRK") || frame.frameId.Equals("TRCK")) {
                    string[] t = frame.GetAsAsciiString().Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
                    if (t.Length > 0)
                        TTools.StringToInteger(ref record.track, t[0]);
                    if (t.Length > 1)
                        TTools.StringToInteger(ref record.trackMax, t[1]);
                }
                else {
                    record.other.Add(frame);
                }
            }
            return true;
        }

        public byte[] MakeTag() {
            if (record.version <= TTagVersion.V_1_1) {
		        return MakeV1Tag();
	        }
	        else {
		        return MakeV2Tag();
	        }
        }

        byte[] FillSpace(byte[] inData, int max) {
            byte[] result = new byte[max];
            for (int i = 0; i < inData.Length; i++) {
                result[i] = inData[i];
            }
            for (int i = inData.Length; i < max; i++) {
                result[i] = 0x20;
            }
            return result;
        }

        public byte[] MakeV1Tag() {
            record.EncodeStrings();
            record.year = TTagRecord.CorrectYear(record.year);

            TTagString title = record.title.Clone();
            TTagString artist = record.artist.Clone();
            TTagString album = record.album.Clone();
            TTagString comment = record.comment.Clone();
            title.EncodeString(TTagEncoding.ANSI, 30);
            artist.EncodeString(TTagEncoding.ANSI, 30);
            album.EncodeString(TTagEncoding.ANSI, 30);
            comment.EncodeString(TTagEncoding.ANSI, record.version >= TTagVersion.V_1_1 ? 28 : 30);

            MemoryStream ms = new MemoryStream();
            BinaryWriter bw = new BinaryWriter(ms);
            ASCIIEncoding ae = new ASCIIEncoding();
            bw.Write(ae.GetBytes("TAG"));
            bw.Write(FillSpace(title.original, 30));
            bw.Write(FillSpace(artist.original, 30));
            bw.Write(FillSpace(album.original, 30));
            bw.Write(FillSpace(ae.GetBytes(record.year), 4));
            if (record.version >= TTagVersion.V_1_1) {
                bw.Write(FillSpace(comment.original, 28));
                bw.Write((byte)0);
                bw.Write((byte)record.track);
            }
            else {
                bw.Write(FillSpace(comment.original, 30));
            }
            bw.Write((byte)record.genre);

            byte[] result = ms.ToArray();
            bw.Close();
            ms.Close();
            return result;
        }

        public byte[] MakeV2Tag() {
            record.EncodeStrings();
            record.year = TTagRecord.CorrectYear(record.year);

            MemoryStream ms = new MemoryStream();
            BinaryWriter bw = new BinaryWriter(ms);

            TID3V2Head head = new TID3V2Head();
            head.major = (byte)record.version;
            head.Write(bw);

        	Boolean isV22 = (record.version == TTagVersion.V_2_2);
            TID3V2Frame frame = new TID3V2Frame(record.version);
            frame.WriteTextFrame(bw, isV22 ? "TT2" : "TIT2", record.title, record.version);
            frame.WriteTextFrame(bw, isV22 ? "TP1" : "TPE1", record.artist, record.version);
            frame.WriteTextFrame(bw, isV22 ? "TAL" : "TALB", record.album, record.version);
            frame.WriteAsciiFrame(bw, isV22 ? "TCO" : "TCON", record.GetGenreExportString(), record.version);
            //////////////////////////////////
            // 
            // 2008.07.04 Takeshi Yoneki
            //frame.WriteAsciiFrame(bw, isV22 ? "TCO" : "TCON", "(100)", m_record.m_version);
            //////////////////////////////////
            frame.WriteAsciiFrame(bw, isV22 ? "TRK" : "TRCK", record.GetTrackAndTrackMaxString(), record.version);
            if (record.year.Length > 0)
                frame.WriteAsciiFrame(bw, isV22 ? "TYE" : "TYER", record.year, record.version);

            foreach (TID3V2Frame c in record.systemComment) {
                if (c.m_version == record.version)
                    c.Write(bw);
            }
            frame.WriteCommentFrame(bw, isV22 ? "COM" : "COMM", record.comment, record.version);
            frame.WritePictureFrame(bw, isV22 ? "PIC" : "APIC", record.picture, record.version);

            foreach (TID3V2Frame f in record.otherPicture) {
                if (f.m_version == record.version)
                    f.Write(bw);
            }

            foreach (TID3V2Frame f in record.other) {
                if (f.m_version == record.version)
                    f.Write(bw);
            }

            long pos = bw.Seek(0, SeekOrigin.Current);
            long sz = 0;
            while (sz < pos + 64) {
                sz += 64;
            }
            long fill = sz - pos;
            for (long i = 0; i < fill; i++)
                bw.Write((byte)0);

            bw.Seek(0, SeekOrigin.Begin);
            head.size = (int)sz;
            head.Write(bw);

            byte[] result = ms.ToArray();
            bw.Close();
            ms.Close();
            return result;
        }

        public void CopyToUndoRecord() {
            undoRecord = record.Clone();
        }

        public void CopyToRedoRecord() {
            redoRecord = record.Clone();
        }

        public void CopyFromUndoRecord() {
            record = undoRecord;
        }

        public void CopyFromRedoRecord() {
            record = redoRecord;
        }
    }

    public class TTagComparer : IComparer<TTagItem> {
        public int Compare(TTagItem x, TTagItem y) {
            int result = 0;
            if (x.container.sortMethod.Equals("file-name")) {
                result = x.record.GetFileName().CompareTo(y.record.GetFileName());
            }
            else if (x.container.sortMethod.Equals("title")) {
                result = x.record.GetTitle().CompareTo(y.record.GetTitle());
            }
            else if (x.container.sortMethod.Equals("album")) {
                result = x.record.GetAlbum().CompareTo(y.record.GetAlbum());
            }
            else if (x.container.sortMethod.Equals("artist")) {
                result = x.record.GetArtist().CompareTo(y.record.GetArtist());
            }
            else if (x.container.sortMethod.Equals("comment")) {
                result = x.record.GetComment().CompareTo(y.record.GetComment());
            }
            else if (x.container.sortMethod.Equals("genre")) {
                if (x.record.genre < y.record.genre)
                    result = -1;
                else if (x.record.genre > y.record.genre)
                    result = 1;
            }
            else if (x.container.sortMethod.Equals("year")) {
                result = x.record.GetYear().CompareTo(y.record.GetYear());
            }
            else if (x.container.sortMethod.Equals("track")) {
                if (x.record.track < y.record.track)
                    result = -1;
                else if (x.record.track > y.record.track)
                    result = 1;
            }
            else if (x.container.sortMethod.Equals("version")) {
                if (x.record.version < y.record.version)
                    result = -1;
                else if (x.record.version > y.record.version)
                    result = 1;
            }
            else if (x.container.sortMethod.Equals("load-encode")) {
                if (x.record.loadEnc < y.record.loadEnc)
                    result = -1;
                else if (x.record.loadEnc > y.record.loadEnc)
                    result = 1;
            }
            else if (x.container.sortMethod.Equals("save-encode")) {
                if (x.record.saveEnc < y.record.saveEnc)
                    result = -1;
                else if (x.record.saveEnc > y.record.saveEnc)
                    result = 1;
            }

            if (result != 0)
                return result;

            result = x.record.GetFileName().CompareTo(y.record.GetFileName());
            if (result == 0)
                result = x.record.GetTitle().CompareTo(y.record.GetTitle());
            if (result == 0) {
                if (x.record.track < y.record.track)
                    result = -1;
                else if (x.record.track > y.record.track)
                    result = 1;
            }
            if (result == 0)
                result = x.record.GetAlbum().CompareTo(y.record.GetAlbum());
            if (result == 0)
                result = x.record.GetArtist().CompareTo(y.record.GetArtist());

            return result;
        }
    }

    public class TTagManager {
        public List<TTagItem> tags = new List<TTagItem>();
        public string sortMethod = "track";
		public ProgressHandler ProgressCallback = null;
        public SaveEndHandler SaveEndCallback = null;
        public RedrawHandler RedrawCallback = null;
        public WorkingHandler WorkingCallback = null;

        Object cueLock = new Object();
        Thread th = null;
        bool working = false;
        bool stopRequest = false;

        TTagItem GetByFilePath(string path) {
            foreach (TTagItem t in tags) {
                if (t.filePath.Equals(path, StringComparison.CurrentCultureIgnoreCase))
                    return t;
            }
            return null;
        }
        
        public int AppendFiles(string[] pathArray) {
            int result = 0;
            if (pathArray == null || pathArray.Length == 0)
                return result;
            foreach (string path in pathArray) {
                result += LoadTagFiles(path);
            }
            return result;
        }

        public int LoadTagFiles(string path) {
            int result = 0;
            if (Directory.Exists(path)) {
                DirectoryInfo info = new DirectoryInfo(path);
                FileInfo[] files = info.GetFiles();
                foreach (FileInfo file in files) {
                    string ext = Path.GetExtension(file.FullName).ToLower();
                    if (ext.Equals(".mp3")) {
                        if (GetByFilePath(file.FullName) == null) {
                            TTagItem t = new TTagItem(this);
                            if (t.LoadTag(file.FullName)) {
                                tags.Add(t);
                                result++;
                            }
                        }
                    }
                }
                Program.thePref.loadDir = path;
            }
            else if (File.Exists(path)) {
                FileInfo info = new FileInfo(path);
                string ext = Path.GetExtension(path).ToLower();
                if (ext.Equals(".mp3")) {
                    if (GetByFilePath(path) == null) {
                        TTagItem t = new TTagItem(this);
                        if (t.LoadTag(path)) {
                            tags.Add(t);
                            result++;
                        }
                        Program.thePref.loadDir = info.DirectoryName;
                    }
                }
            }
            return result;
        }

        public bool IsWorking() {
            lock (cueLock) {
                return working;
            }
        }

        public void Stop() {
            lock (cueLock) {
                stopRequest = true;
            }
            if (th != null) {
                th.Join();
            }
            foreach (TTagItem tag in tags) {
                if (tag.tempFilePath.Length > 0 && File.Exists(tag.tempFilePath)) {
                    File.Delete(tag.tempFilePath);
                }
            }
            if (ProgressCallback != null) {
                ProgressCallback(0);
            }
            if (RedrawCallback != null) {
                RedrawCallback(tags);
            }
        }

        private bool IsContinue() {
            lock (cueLock) {
                return !stopRequest;
            }
        }

        public void SaveFiles() {
            if (WorkingCallback()) {
                return;
            }
            int count = tags.Count;
            if (count == 0) {
                return;
            }
            stopRequest = false;
            th = new Thread(new ThreadStart(SaveFilesThread));
            working = true;
            th.Start();
        }

        public void SaveFilesThread() {
            int count = tags.Count;
            bool[] bSaveTemp = new bool[count];
            bool keepFileTime = false;

            DateTime[] lastWriteTime = new DateTime[count];
            DateTime[] creationTime = new DateTime[count];

            long soundSizeAll = 0;
            for (int index = 0; index < count; index++) {
                TTagItem tag = tags[index];
                soundSizeAll += tag.soundSize;
            }
            long soundSizeDone = 0;
            int percent = 0;

            for (int index = 0; IsContinue() && index < count; index++) {
                bSaveTemp[index] = false;
                TTagItem tag = tags[index];
                if (tag.record.bModify == false)
                    continue;
                if (tag.record.version >= TTagVersion.V_2_2)
                    bSaveTemp[index] = true;
                if (tag.record.fileName.Equals(Path.GetFileName(tag.filePath)) == false)
                    bSaveTemp[index] = true;
                if (tag.soundStartPos != 0)
                    bSaveTemp[index] = true;

                FileInfo fInfo = new FileInfo(tag.filePath);
                lastWriteTime[index] = fInfo.LastWriteTime;
                creationTime[index] = fInfo.CreationTime;

                if (bSaveTemp[index] == false) {
                    byte[] tag1Data = tag.MakeV1Tag();
                    if (tag1Data.Length == 0)
                        continue;
                    // V1.x only
                    FileStream fsOut = new FileStream(tag.filePath, FileMode.Open, FileAccess.ReadWrite);
                    BinaryWriter bw = new BinaryWriter(fsOut);
                    bw.BaseStream.SetLength(tag.soundSize);
                    bw.BaseStream.Seek(0, SeekOrigin.End);
                    bw.Write(tag1Data);
                    bw.Close();
                    fsOut.Close();
                    if (keepFileTime) {
                        fInfo.LastWriteTime = lastWriteTime[index];
                        fInfo.CreationTime = creationTime[index];
                    }
                    soundSizeDone += tag.soundSize;
                    percent = (int)(100 * soundSizeDone / soundSizeAll);
                    Progress(percent);
                }
                else {
                    byte[] tagData = tag.MakeTag();
                    if (tagData.Length == 0)
                        continue;
                    // V1.x or V2.x
                    tag.tempFilePath = tag.filePath + ".~temp~";
                    FileStream fsIn = new FileStream(tag.filePath, FileMode.Open, FileAccess.Read);
                    BufferedStream br = new BufferedStream(fsIn, 64 * 1024);
                    FileStream fsOut = new FileStream(tag.tempFilePath, FileMode.Create, FileAccess.Write);
                    BufferedStream bsOut = new BufferedStream(fsOut, 64 * 1024);
                    BinaryWriter bw = new BinaryWriter(bsOut);

                    if (tag.record.version >= TTagVersion.V_2_2)
                        bw.Write(tagData);
                    br.Seek(tag.soundStartPos, SeekOrigin.Begin);
                    for (int i = 0; i < tag.soundSize; i++) {
                        bw.Write((byte)br.ReadByte());
                        soundSizeDone++;
                        int p = (int)(100 * soundSizeDone / soundSizeAll);
                        if (p != percent) {
                            percent = p;
                            Progress(percent);
                        }
                    }
                    // V1^OóB
                    // ߂B
                    //bw.Write(tag.MakeV1Tag());
                    if (tag.record.version < TTagVersion.V_2_2)
                        bw.Write(tagData);
                    if (tag.record.version < TTagVersion.V_2_2)
                        tag.soundStartPos = 0;
                    else
                        tag.soundStartPos = tagData.Length;
                    bw.Close();
                    bsOut.Close();
                    fsOut.Close();
                    br.Close();
                    fsIn.Close();
                }
            }

            if (IsContinue()) {
                for (int index = 0; index < count; index++) {
                    TTagItem tag = tags[index];
                    if (tag.record.bModify) {
                        if (bSaveTemp[index])
                            File.Delete(tag.filePath);
                    }
                }

                for (int index = 0; index < count; index++) {
                    TTagItem tag = tags[index];
                    if (tag.record.bModify) {
                        string path = Path.Combine(Path.GetDirectoryName(tag.filePath), tag.record.fileName);
                        if (bSaveTemp[index]) {
                            try {
                                File.Move(tag.tempFilePath, path);
                            }
                            catch (Exception e) {
                            }
                        }
                        tag.filePath = path;
                        // t@CXVێB
                        if (keepFileTime) {
                            FileInfo fInfo = new FileInfo(tag.filePath);
                            fInfo.LastWriteTime = lastWriteTime[index];
                            fInfo.CreationTime = creationTime[index];
                        }
                    }
                    tag.record.loadEnc = tag.record.saveEnc;
                    tag.record.bModify = false;

                    // \̍XVB
                    //if (MainViewRedrawCallback != null) {
                    //    List<TTagItem> tags = new List<TTagItem>();
                    //    tags.Add(tag);
                    //    MainViewRedrawCallbackThread(tags);
                    //}
                }
            }
            Redraw(tags);

            SaveEnd();
            working = false;
        }

        public bool Export(string filePath, string ext) {
            FileStream fs = new FileStream(filePath, FileMode.Create, FileAccess.Write);
            if (fs == null)
                return false;
            StreamWriter sw = new StreamWriter(fs);
            foreach (TTagItem t in tags) {
                string s = "";
                if (ext.Equals(".tsv"))
                    s = t.record.GetTSVLine();
                else if (ext.Equals(".csv"))
                    s = t.record.GetCSVLine();
                sw.WriteLine(s);
            }
            sw.Close();
            fs.Close();
            return true;
        }

        public bool Import(string filePath, string ext) {
            FileStream fs = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            if (fs == null)
                return false;
            StreamReader sr = new StreamReader(fs);
            List<string> lines = new List<string>();
            while (true) {
                string s = sr.ReadLine();
                if (s == null || s.Length == 0)
                    break;
                lines.Add(s);
            }
            sr.Close();
            fs.Close();
            for (int i = 0; i < tags.Count && i < lines.Count; i++) {
                if (ext.Equals(".tsv"))
                    tags[i].record.SetTSVLine(lines[i]);
                else if (ext.Equals(".csv"))
                    tags[i].record.SetCSVLine(lines[i]);
            }
            return true;
        }

        public bool IsModified() {
            foreach (TTagItem t in tags) {
                if (t.record.bModify)
                    return true;
            }
            return false;
        }

        void Progress(int percent) {
            // ~vJoinĂ΂邽߁AUIXbh͑҂ɂȂB
            // Ȃ̂Invoke܂ތĂяo̓fbhbNB
            if (IsContinue() && ProgressCallback != null) {
                ProgressCallback(percent);
            }
        }

        void SaveEnd() {
            if (IsContinue() && SaveEndCallback != null) {
                SaveEndCallback();
            }
        }

        void Redraw(List<TTagItem> tags) {
            if (IsContinue() && RedrawCallback != null) {
                RedrawCallback(tags);
            }
        }

    } // class
} // namespace
